# -*- coding: utf-8 -*-
"""
keri.core.serdering module

"""
import copy
import json
from collections import namedtuple
from collections.abc import Mapping
from dataclasses import dataclass, asdict, field

import cbor2 as cbor
import msgpack
import pysodium
import blake3
import hashlib
from  ordered_set import OrderedSet as oset


from .. import kering
from ..kering import (ValidationError,  MissingFieldError, ExtraFieldError,
                      AlternateFieldError, InvalidValueError,
                      ShortageError, VersionError, ProtocolError, KindError,
                      DeserializeError, FieldError, SerializeError)
from ..kering import (Versionage, Version, Vrsn_1_0, Vrsn_2_0,
                      VERRAWSIZE, VERFMT,
                      MAXVERFULLSPAN, VER1FULLSPAN,  VER2FULLSPAN)
from ..kering import SMELLSIZE, Smellage, smell

from ..kering import Protocols, Kinds, versify, deversify, Ilks

from .. import help
from ..help import helping


from . import coring
from .coring import (MtrDex, DigDex, PreDex, NonTransDex, PreNonDigDex,
                     Saids,  Digestage)
from .coring import (Matter, Saider, Verfer, Diger, Number, Tholder, Tagger,
                     Ilker, Traitor, Verser, )

from .counting import GenDex, Counter, Codens, SealDex_2_0

from .structing import Sealer, SClanDom



logger = help.ogler.getLogger()




@dataclass
class FieldDom:
    """
    Field configuration dataclass for Serder messages. Provides field labels
    and default field values for a given ilk (message type).

    Attributes:
        alls (dict): Allowed fields (not extra)
                    alls must not be empty since at least version string
                    or protocol version filed is always required.
                    Fields in alls that appear must appear in order.

        opts (dict): Optional fields within alls.
                    opts defaults to empty.
                    When opts is empty than all alls are required.
                    Any fields in alls but not in opts are required.
                    opts is a subset of alls

        alts (dict): Alternate fields within alls.
                    alts defaults to empty.
                    An alt field means that one or the other of two fields is
                    allowed but not both. Two entries in alts are required, one
                    for each field in an alt pair. the alt pair key value are
                    the two labels. One entry for each order.
                    all alts must be in opts since both can't be required.
                    Suppose 'a' and 'A' are an alternate pair then alts =
                    { 'a': 'A', 'A': 'a'}. This allows the presence of one to
                    block the presence of the other by looking up the one present
                    as key to see the value of the one to block.

        saids (dict): are saidive fields whose value may be computed as a said of the message.
                    saids defaults to empty
                    when provided a field in saids indicates the field value is saidive.
                    A simple SAID field value is always computed.
                    An AID SAID field value is only computed when its code indicates.
                    saids is a subset of alls

        strict (bool): determines if alls is strict, no extra fields are allowed
                       strict defaults to True.
                       True means no extra fields are allowed, only those in alls.
                       False means extra fields are allowed besides thos in alls.
                       Extra fields are fields not in alls.
                       Extra fields may appear in any order after the last field
                       in alls.


        When strict:  no extras allowed
           Any fields not in alls raise error
           If opts is empty then all alls are required in order
           If opts is not empty then fields in opts are optional but the rest of
                the fields in alls are required

        When not strict: extras allowed
           Any fields not in alls must appear after all fields in alls
           If opts is empty then all alls are required in order
           If opts is not empty then fields in opts are optional but the rest of
                the fields in alls are required

    """
    alls: dict  # all allowed fields when strict
    opts: dict = field(default_factory=dict)  # optional fields
    alts:  dict = field(default_factory=dict)  # alternate optional fields
    saids: dict = field(default_factory=dict)  # saidive fields
    strict: bool = True  # only alls allowed no extras

    def __iter__(self):
        return iter(asdict(self))

"""Design Notes:

Problems is that sniff only determines if counter not which type of
counter. Then smell does regex lookahead to find out which serialization
when not count code but extractor does not look ahead but strips from
stream. So when possibility that CESR message is next either need to
not strip from stream when extracting or if counter is message
then grab rest of frame and reattach so raw in Serder includes the
message counter. Latter is better since always keep counter around
until later. So need to check counter type and if message then
extract rest of counter frame (message) and reattach counter raw.
Then can call Serder with raw and smellage that indicates CESR kind

But this does not solve the problem of using the Serder subclass
for the given protocol. Merely knowing is a CESR message is not
enough also have to know the protocol which comes in the version
field (not version string).

One solution is to modify smell so that it also can lookahead and
see the version field. Or lookahead and see the version field with
count codes in front. Problem is that the Regexes don't separate
cleanly.

Another solution is to use distinct function for cesr native called
snuff like smell but regex only for CESR native. Reap can be told which
because sniff tells which it is.
So question for snuff is should it be searching over the counter or should it
start at version field. This changes regex so forced start at front of raw.
so if reattach counter but use skip then can snuff at start of string.
Begine regex with b'^' or b'\A' to match at start of string.

So change Smellage to return extra field that has gvrsn when used by snuff
so can use Smellage for both smell and snuff in both reap and inhale
where smellage is used. Change egacy uses of smell to ignore extra value.



Lets try that as it works the best.

while True:  # extract, deserialize, and strip message from ims
    try:
        serder = serdery.reap(ims=ims)  # can set version here
    except kering.ShortageError as ex:  # need more bytes
        yield
    else: # extracted and stripped successfully
        break  # break out of while loop

ctr = yield from self._extractor(ims=ims, klas=Counter, cold=cold)
if ctr.code == CtrDex.AttachmentGroup:  # pipeline ctr?
    pipelined = True

@staticmethod
def _extractor(ims, klas, cold=Colds.txt, abort=False):
    while True:
        try:
            if cold == Colds.txt:
                return klas(qb64b=ims, strip=True)
            elif cold == Colds.bny:
                return klas(qb2=ims, strip=True)
            else:
                raise kering.ColdStartError("Invalid stream state cold={}.".format(cold))
        except kering.ShortageError as ex:
            if abort:  # pipelined pre-collects full frame before extracting
                raise  # bad pipelined frame so abort by raising error
            yield



"""


class Serdery:
    """Serder factory class for generating serder instances by protocol type
    from an incoming message stream.
    """

    def __init__(self, *pa, **kwa):
        """Init instance

        Parameters:

        """
        pass


    def reap(self, ims, genus=GenDex.KERI, gvrsn=Vrsn_2_0, native=False, skip=0):
        """Extract and return Serder subclass based on protocol type reaped from
        version string inside serialized raw of Serder.

        Returns:
            serder (Serder): instance of Serder subclass where subclass is
                determined by the protocol type of its version string.

        Parameters:
            ims (bytearray) of serialized incoming message stream. Assumes start
                of stream is raw Serder.
            genus (str): CESR genus code from stream parser.
                    Provides genus of enclosing stream top-level or nested group
            gvrsn (Versionage): instance CESR genus code table version (Major, Minor)
                    Provides genus of enclosing stream top-level or nested group
            native (bool): True means sniff determined may be CESR native message
                           so snuff instead of smell.
                           False means sniff determined not CESR native i.e
                           JSON, CBOR, MGPK field map. so use smell. Default False
             skip (int): bytes to skip at front of ims. Useful when CESR native
                serialization where skip is size of the message counter so smell
                does need to see counter

        """
        if native:
            pass
            #smellage = smell(memoryview(ims)[skip:])  # does not copy to skip

        else:
            smellage = smell(ims)

        if smellage.proto == Protocols.keri:
            return SerderKERI(raw=ims, strip=True, smellage=smellage,
                              genus=genus, gvrsn=gvrsn)
        elif smellage.proto == Protocols.acdc:
            return SerderACDC(raw=ims, strip=True, smellage=smellage,
                              genus=genus, gvrsn=gvrsn)
        else:
            raise ProtocolError(f"Unsupported protocol type = {smellage.proto}.")



class Serder:
    """Serder is serializer-deserializer class for saidified  over-the-wire
    messages that deserializes to a field map (label value pairs) from
    either a serialized field map or an unlabeled fixed field structure with
    affiliated label list. The messages must include a version string field
    with proto (protocol), version, kind, and size elements or equivalent
    header with same elements. The messages may have an optional ilk field
    that is protocol specific. Protocols that have fixed  top-level fields
    also perform label inclusion validation.

    Message saidification and verification may be dependent on protocol and
    optionally ilk specific field label(s) and digest code type for its SAID(s).

    The base Serder class provides the common properties for all messages for
    all protocols. Each subclass is protocol based and adds properties that are
    required for all message ilks in a given protocol. Each protocol subclass
    may have dynamically injected ilk specific properties (as descriptors)
    if any.

    To support a new protocol with its ilks, add a protocol specific subclass,
    override the Labels class attribute, and if necessary the .verify and
    .saidify methods and define any protocol specific properties. If necessary
    define ilk specific subclasses of a given protocol (ilk is packet type).
    The .Labels class attributes configures the field label(s) for said
    generation and verification in addition to the required fields.

    Class Attributes:
        Dummy (str): dummy character for computing SAIDs
        Spans (dict): version string spans keyed by version
        Digests (dict): map of digestive codes. Should be same set of codes as
            in coring.DigestCodex coring.DigDex so that .digestive property works.
            Use unit tests to ensure codex sets match
        Protocol (str): class specific message protocol
        Proto (str): default message protocol
        Vrsn (Versionage): default version
        Kind (str): default serialization kind one of Serials
        Fields (dict): nested dict of field labels keyed by protocol, version,
            and message type (ilk). Felds labels are provided with a Fieldage
            named tuple (saids, reqs, alls) that governs field type and presence.
            None is default message type (ilk) when no ilk needed in a message.
            See below for detailed logic associated with Fields class attribute

    Properties:
        raw (bytes): of serialized event only
        sad (dict): self addressed data dict
        genus (str): CESR genus code
        gvrsn (Versionage): instance CESR genus code table version (Major, Minor)
        proto (str): Protocolage value as protocol identifier such as KERI, ACDC
                     alias of .protocol
        protocol (str): Protocolage value as protocol identifier such as KERI, ACDC
                        alias of .proto
        vrsn (Versionage): protocol version (Major, Minor) alias of .version
        version (Versionage): protocol version (Major, Minor) alias of .vrsn
        kind (str): serialization kind coring.Serials such as JSON, CBOR, MGPK, CESR
        size (int): number of bytes in serialization
        said (str): qb64 said of .raw given by appropriate field
        saidb (bytes): qb64b of .said
        ilk (str | None): packet type for this Serder if any (may be None)


    Hidden Attributes:
        ._raw (bytes): serialized message
        ._sad (dict): sad dict (key event dict)
        ._cvrsn (Versionage): CESR code table version
        ._proto (str):  Protocolage value as protocol type identifier
        ._vrsn is Versionage instance of event version
        ._kind is serialization kind string value (see namedtuple coring.Serials)
            supported kinds are 'json', 'cbor', 'msgpack', 'binary'
        ._size is int of number of bytes in serialed event only
        ._said (str): qb64 given by appropriate saidive field

    Methods:
        verify()
        _verify()
        makify()
        compare()
        pretty(size: int | None ) -> str: Prettified JSON of this SAD

    ClassMethods:
        _inhale()
        _exhale()

    StaticMethods:
        loads()
        dumps()

    Note:
        loads and jumps of json use str whereas cbor and msgpack use bytes


    Fields:
        Each element of Fields is a FieldDom dataclass instance with four attributes:
            alls (dict):
            opts (dict):
            saids (dict):
            strict (bool):

    """
    Dummy = "#"  # dummy spaceholder char for SAID. Must not be a valid Base64 char

    # Spans dict keyed by version (Versionage instance) of version string span (size)
    Spans = {Vrsn_1_0: VER1FULLSPAN, Vrsn_2_0: VER2FULLSPAN}

    # map seal clan names to seal counter code for grouping seals in anchor list
    ClanCodes = dict()
    ClanCodes[SClanDom.SealDigest.__name__] = SealDex_2_0.DigestSealSingles
    ClanCodes[SClanDom.SealRoot.__name__] = SealDex_2_0.MerkleRootSealSingles
    ClanCodes[SClanDom.SealBacker.__name__] = SealDex_2_0.BackerRegistrarSealCouples
    ClanCodes[SClanDom.SealLast.__name__] = SealDex_2_0.SealSourceLastSingles
    ClanCodes[SClanDom.SealTrans.__name__] = SealDex_2_0.SealSourceCouples
    ClanCodes[SClanDom.SealEvent.__name__] = SealDex_2_0.SealSourceTriples

    # map seal counter code to seal clan name for parsing seal groups in anchor list
    CodeClans = { val: key for key, val in ClanCodes.items()}  # invert dict

    #override in subclass to enforce specific protocol
    Protocol = None  # class based message protocol, None means any in Protocols is ok
    Proto = Protocols.keri  # default message protocol type for makify on base Serder
    Vrsn = Vrsn_1_0  # default protocol version for protocol type
    Kind = Kinds.json  # default serialization kind
    CVrsn = Vrsn_2_0  # default CESR code table version


    # Nested dict keyed by protocol.
    # Each protocol value is a dict keyed by ilk.
    # Each ilk value is a Labelage named tuple with saids, codes and fields
    # ilk value of None is default for protocols that support ilkless packets
    Fields = \
    {
        Protocols.keri:
        {
            Vrsn_1_0:
            {
                Ilks.icp: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        kt='0',k=[], nt='0', n=[], bt='0', b=[], c=[],
                        a=[]),
                    saids={Saids.d: DigDex.Blake3_256,
                           Saids.i: DigDex.Blake3_256,}),
                Ilks.rot: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                            p='', kt='0',k=[], nt='0', n=[], bt='0', br=[],
                            ba=[], a=[]),
                        saids={Saids.d: DigDex.Blake3_256}),
                Ilks.ixn: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        p='', a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.dip: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        kt='0', k=[], nt='0', n=[], bt='0', b=[], c=[],
                        a=[], di=''),
                    saids={Saids.d: DigDex.Blake3_256,
                           Saids.i: DigDex.Blake3_256,}),
                Ilks.drt: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        p='', kt='0',k=[], nt='0', n=[], bt='0', br=[],
                        ba=[], a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.rct: FieldDom(alls=dict(v='', t='',d='', i='', s='0')),
                Ilks.qry: FieldDom(alls=dict(v='', t='',d='', dt='', r='',
                        rr='',q={}),
                    saids={Saids.d: DigDex.Blake3_256},),
                Ilks.rpy: FieldDom(alls=dict(v='', t='',d='', dt='', r='',
                        a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.pro: FieldDom(alls=dict(v='', t='',d='', dt='', r='',
                        rr='',q={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.bar: FieldDom(alls=dict(v='', t='',d='', dt='', r='',
                        a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.exn: FieldDom(alls=dict(v='', t='', d='', i="", rp="",
                        p="", dt='', r='',q={}, a=[], e={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.vcp: FieldDom(alls=dict(v='', t='',d='', i='', ii='',
                        s='0', c=[], bt='0', b=[], n=''),
                    saids={Saids.d: DigDex.Blake3_256,
                           Saids.i: DigDex.Blake3_256,}),
                Ilks.vrt: FieldDom(alls=dict(v='', t='',d='', i='', p='',
                        s='0', bt='0', br=[], ba=[]),
                    saids={Saids.d: DigDex.Blake3_256,}),
                Ilks.iss: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        ri='', dt=''),
                    saids={Saids.d: DigDex.Blake3_256,}),
                Ilks.rev: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        ri='', p='', dt=''),
                    saids={Saids.d: DigDex.Blake3_256,}),
                Ilks.bis: FieldDom(alls=dict(v='', t='',d='', i='', ii='',
                        s='0', ra={}, dt=''),
                    saids={Saids.d: DigDex.Blake3_256,}),
                Ilks.brv: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                         p='', ra={}, dt=''),
                    saids={Saids.d: DigDex.Blake3_256,}),
            },
            Vrsn_2_0:
            {
                Ilks.icp: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                         kt='0', k=[], nt='0', n=[], bt='0', b=[], c=[],
                         a=[]),
                    saids={Saids.d: DigDex.Blake3_256,
                           Saids.i: DigDex.Blake3_256,}),
                Ilks.rot: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        p='', kt='0',k=[], nt='0', n=[], bt='0', br=[],
                        ba=[], c=[], a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.ixn: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        p='', a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.dip: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                        kt='0', k=[], nt='0', n=[], bt='0', b=[], c=[],
                        a=[], di=''),
                    saids={Saids.d: DigDex.Blake3_256,
                           Saids.i: DigDex.Blake3_256,}),
                Ilks.drt: FieldDom(alls=dict(v='', t='',d='', i='', s='0',
                         p='', kt='0',k=[], nt='0', n=[], bt='0', br=[],
                        ba=[], c=[], a=[]),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.rct: FieldDom(alls=dict(v='', t='',d='', i='', s='0')),
                Ilks.qry: FieldDom(alls=dict(v='', t='',d='', i='', dt='',
                        r='', rr='', q={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.rpy: FieldDom(alls=dict(v='', t='',d='', i='', dt='',
                        r='',a={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.pro: FieldDom(alls=dict(v='', t='',d='', i='', dt='',
                        r='', rr='', q={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.bar: FieldDom(alls=dict(v='', t='',d='', i='', dt='',
                        r='',a={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.xip: FieldDom(alls=dict(v='', t='', d='', i="", dt='',
                                             r='', q={}, a={}),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.exn: FieldDom(alls=dict(v='', t='', d='', i="", x="",
                        p="", dt='', r='', q={}, a={}),
                    saids={Saids.d: DigDex.Blake3_256}),
            },
        },
        Protocols.acdc:
        {
            Vrsn_1_0:
            {
                None: FieldDom(alls=dict(v='', d='', u='', i='',
                        ri='', s='', a='', A='', e='', r=''),
                    opts=dict(u='', ri='', a='', A='', e='', r=''),
                    alts=dict(a="A", A="a"),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.ace: FieldDom(alls=dict(v='', t='', d='', u='', i='',
                        ri='', s='', a='', A='', e='', r=''),
                    opts=dict(u='', ri='', a='', A='', e='', r=''),
                    alts=dict(a="A", A="a"),
                    saids={Saids.d: DigDex.Blake3_256},
                    strict=False),
            },
            Vrsn_2_0:
            {
                None: FieldDom(alls=dict(v='', d='', u='', i='',
                        rd='', s='', a='', A='', e='', r=''),
                    opts=dict(u='', rd='', a='', A='', e='', r=''),
                    alts=dict(a="A", A="a"),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.acd: FieldDom(alls=dict(v='', t='', d='', u='', i='',
                        rd='', s='', a='', A='', e='', r=''),
                    opts=dict(u='', rd='', a='', A='', e='', r=''),
                    alts=dict(a="A", A="a"),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.ace: FieldDom(alls=dict(v='', t='', d='', u='', i='',
                        rd='', s='', a='', A='', e='', r=''),
                    opts=dict(u='', rd='', a='', A='', e='', r=''),
                    alts=dict(a="A", A="a"),
                    saids={Saids.d: DigDex.Blake3_256},
                    strict=False),
                Ilks.sch: FieldDom(alls=dict(v='', t='', d='', s=''),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.att: FieldDom(alls=dict(v='', t='', d='', a=''),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.agg: FieldDom(alls=dict(v='', t='', d='', A=''),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.rul: FieldDom(alls=dict(v='', t='', d='', e=''),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.rip: FieldDom(alls=dict(v='', t='', d='', u='', i='',
                                              s='', dt=''),
                    saids={Saids.d: DigDex.Blake3_256}),
                Ilks.upd: FieldDom(alls=dict(v='', t='', d='', r='', s='',
                                              p='', dt='', a=''),
                    saids={Saids.d: DigDex.Blake3_256}),
            },
        },
    }


    def __init__(self, *, raw=b'', sad=None, strip=False, smellage=None,
                 genus=GenDex.KERI, gvrsn=Vrsn_2_0, verify=True, makify=False,
                 proto=None, vrsn=None, kind=None, ilk=None, saids=None):
        """Deserialize raw if provided. Update properties from deserialized raw.
            Verifies said(s) embedded in sad as given by labels.
            When verify is True then verify said(s) in deserialized raw as
            given by label(s) according to proto and ilk and code
        If raw not provided then serialize .raw from sad with kind and code.
            When kind not provided use kind embedded in sad['v'] version string.
            When saidify is True then compute and update said(s) in sad as
            given by label(s) according to proto and ilk and code.

        Parameters:
            raw (bytes | bytearray): serialized event
            sad (dict): serializable saidified field map of message.
                Ignored if raw provided
            strip (bool): True means strip (delete) raw from input stream
                bytearray after parsing. False means do not strip.
                Assumes that raw is bytearray when strip is True.
            smellage (Smellage | None): instance of deconstructed and converted
                version string elements. If none or empty ignore otherwise assume
                that raw already had its version string extracted (reaped) into the
                elements of smellage.
            genus (str):  CESR genus str. Either provided by parser from stream
                or generated by Serder to stream
            gvrsn (Versionage): instance CESR genus code table version
                Either provided by parser from stream genus version or desired when
                generating Serder instance to stream
            verify (bool): True means verify said(s) of given raw or sad.
                           False means don't verify. Useful to avoid unnecessary
                           reverification when deserializing from database
                           as opposed to over the wire reception.
                           Raises ValidationError if verification fails
                           Ignore when raw empty or when raw and saidify is True
            makify (bool): True means compute fields for sad including size and
                saids.
            proto (str | None): desired protocol type str value of Protocols
                If None then its extracted from sad or uses default .Proto
            vrsn (Versionage | None): instance desired protocol version
                If None then its extracted from sad or uses default .Vrsn
            kind (str None): serialization kind string value of Serials
                supported kinds are 'json', 'cbor', 'msgpack', 'binary'
                If None then its extracted from sad or uses default .Kind
            ilk (str | None): desired ilk packet type str value of Ilks
                If None then its extracted from sad or uses default .Ilk
            saids (dict): of keyed by label of codes for saidive fields to
                override defaults given in .Fields for a given ilk.
                If None then use defaults
                Code assignment for each saidive field in desending priority:
                   - the code provided in saids when not None
                   - the code extracted from sad[said label] when valid CESR
                   - the code provided in .Fields...saids


        """
        self._gvrsn = gvrsn
        self._genus = genus

        if raw:  # deserialize raw using property setter
            self._inhale(raw=raw, smellage=smellage)
            # ._inhale updates ._raw, ._sad, ._proto, ._vrsn, ._kind, ._size

            # primary said field label
            try:
                label = list(self.Fields[self.proto][self.vrsn][self.ilk].saids)[0]
                if label not in self._sad:
                    raise FieldError(f"Missing primary said field in {self._sad}.")
                self._said = self._sad[label]  # not verified
            except Exception as ex:
                self._said = None  # no saidive field

            if strip:  #only when raw is bytearray
                try:
                    del raw[:self._size]
                except TypeError:
                    pass  # ignore if bytes

            if verify:  # verify fields including the said(s) provided in raw
                try:
                    self._verify()  # raises exception when not verify
                except Exception as ex:
                    logger.error("Invalid raw for Serder %s\n%s",
                                 self.pretty(), ex.args[0])
                    raise ValidationError(f"Invalid raw for Serder = "
                                          f"{self._sad}. {ex.args[0]}") from ex

        elif sad or makify:  # serialize sad into raw or make sad
            if makify:  # recompute properties and said(s) and reset sad
                # makify resets sad, raw, proto, vrsn, kind, ilk, and size
                self.makify(sad=sad, proto=proto, vrsn=vrsn, kind=kind,
                            ilk=ilk, saids=saids)
                # .makify updates ._raw, ._sad, ._proto, ._vrsn, ._kind, ._size

            else:
                self._exhale(sad=sad)
                # .exhale updates ._raw, ._sad, ._proto, ._vrsn, ._kind, ._size

            # primary said field label
            try:
                label = list(self.Fields[self.proto][self.vrsn][self.ilk].saids)[0]
                if label not in self._sad:
                    raise DeserializeError(f"Missing primary said field in {self._sad}.")
                self._said = self._sad[label]  # not verified
            except Exception:
                self._said = None  # no saidive field

            if verify:  # verify fields including the said(s) provided in sad
                try:
                    self._verify()  # raises exception when not verify
                except Exception as ex:
                    logger.error("Invalid sad for Serder %s\n%s",
                                 self.pretty(), ex.args[0])
                    raise ValidationError(f"Invalid sad for Serder ="
                                              f"{self._sad}.") from ex

        else:
            raise InvalidValueError("Improper initialization need raw or sad "
                                    f"or makify.")



    def verify(self):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Returns:
            verify (bool): True if said(s) verify. False otherwise
        """
        try:
            self._verify()
        except Exception as ex:              # log validation error here
            logger.error("Invalid Serder: %s\n for %s\n",
                         ex.args[0], self.pretty())

            return False

        return True


    def _verify(self):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        if self.Protocol and self.proto != self.Protocol:  # class required
            raise ValidationError(f"Required protocol = {self.Protocol}, got "
                                 f"{self.proto} instead.")

        if self.proto not in self.Fields:
            raise ValidationError(f"Invalid protocol type = {self.proto}.")

        if self.genus not in GenDex:  # ensures self.genus != None
            raise SerializeError(f"Invalid genus={self.genus}.")

        if getattr(GenDex, self.proto, None) != self.genus:
            raise SerializeError(f"Incompatible protocol={self.proto} with "
                                 f"genus={self.genus}.")

        if self.vrsn.major > self.gvrsn.major:
            raise SerializeError(f"Incompatible major protocol version={self.vrsn}"
                                 f" with major genus version={self.gvrsn}.")

        if self.vrsn not in self.Fields[self.proto]:
            raise SerializeError(f"Invalid version={self.vrsn} for "
                                 f"protocol={self.proto}.")

        if self.ilk not in self.Fields[self.proto][self.vrsn]:
            raise ValidationError(f"Invalid packet type (ilk) = {self.ilk} for"
                                  f"protocol = {self.proto}.")



        fields = self.Fields[self.proto][self.vrsn][self.ilk]  # get labelage

        alls = fields.alls  # faster local reference
        oalls = oset(alls)  # ordereset of field labels
        oopts = oset(fields.opts)  # ordereset of field labels
        oreqs = oalls - oopts  # required fields

        oskeys = oset(self._sad)  # ordered set of keys in sad (skeys)
        osexts = oskeys - oalls  # get ordered set of extras in sad (sexts)
        if osexts and fields.strict:
            raise ExtraFieldError(f"Unallowed extra field(s) = {list(osexts)} "
                                     f"in sad.")

        osopts = oskeys - oreqs - osexts  # subset of opts in sad

        osalls = oalls - (oopts - osopts)  # subset of alls without missing opts in sad

        for k, v in fields.alts.items():
            if k in osopts and v in osopts:
                raise AlternateFieldError(f"Unallowed, alternate fields '{k}' "
                                          f"and '{v}' both present in sad.")

        # can't do set math osalls == oskeys - osexts becasue of osexts might be
        # out-of-order so have to iterate to ensure osexts if any appear in oskeys
        # after all fields in osalls
        for i, label in enumerate(osalls):
            if oskeys[i] != label:
                raise MissingFieldError(f"Missing or out-of-order field = {label} "
                                         f"from = {list(osalls)} in sad.")

        # said field labels are not order dependent with respect to all fields
        # in sad so use set() to test inclusion
        saids = copy.copy(fields.saids)  # get copy of saidive field labels and defaults values
        if not (set(saids) <= set(alls)):
            raise MissingFieldError(f"Missing one or more required said fields"
                                    f" from {list(saids)} in sad = "
                                    f"{self._sad}.")

        sad = self.sad  # make shallow copy so don't clobber original .sad
        for label in saids:
            try:  # replace default code with code of value from sad
                saids[label] = Matter(qb64=sad[label]).code
            except Exception as ex:
                if saids[label] in DigDex:  # digestive but invalid
                    raise ValidationError(f"Invalid said field '{label}' in sad\n"
                                      f" = {self._sad}.") from ex

            if saids[label] in DigDex:  # if digestive then replace with dummy
                sad[label] = self.Dummy * len(sad[label])

        # compute saidive digestive field values using raw from sized dummied sad
        raw = self.dumps(sad, kind=self.kind)  # serialize dummied sad copy
        for label, code in saids.items():
            if code in DigDex:  # subclass override if non digestive allowed
                dig = Diger(ser=raw, code=code).qb64
                if dig != self._sad[label]:  # compare to original
                    raise ValidationError(f"Invalid said field '{label}' in sad"
                                          f" = {self._sad}, should be {dig}.")
                sad[label] = dig

        raw = self.dumps(sad, kind=self.kind)  # compute final raw

        if raw != self.raw:
            raise ValidationError(f"Invalid round trip of {sad} != \n"
                                  f"{self.sad}.")

        if "v" not in sad:
            raise ValidationError(f"Missing version string field in {sad}.")

        # extract version string elements to verify consistency with attributes
        proto, vrsn, kind, size, opt = deversify(sad["v"])
        if self.proto != proto:
            raise ValidationError(f"Inconsistent protocol={self.proto} in {sad}.")

        if self.vrsn != vrsn:
            raise ValidationError(f"Inconsistent version={self.vrsn} in {sad}.")

        if self.kind != kind:
            raise ValidationError(f"Inconsistent kind={self.kind} in {sad}.")

        if self.kind in (Kinds.json, Kinds.cbor, Kinds.mgpk):
            if size != self.size != len(raw):
                raise ValidationError(f"Inconsistent size={self.size} in {sad}.")
        else:  # size is not set in version string when kind is CESR
            if self.size != len(raw):
                raise ValidationError(f"Inconsistent size={self.size} in {sad}.")

        # verified successfully since no exception


    def makify(self, sad, *, proto=None, vrsn=None, kind=None,
               ilk=None, saids=None):
        """Makify given sad dict makes the versions string and computes the said
        field values and sets associated properties:
        raw, sad, proto, version, kind, size

        Override for protocol and ilk specific saidification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Default prioritization.
           Use method parameter if not None
           Else use provided version string if valid
           Otherwise use class attribute


        Parameters:
            sad (dict): serializable saidified field map of message.
                Ignored if raw provided
            proto (str | None): desired protocol type str value of Protocols
                If None then its extracted from sad or uses default .Proto
            vrsn (Versionage | None): instance desired protocol version
                If None then its extracted from sad or uses default .Vrsn
            kind (str None): serialization kind string value of Serials
                supported kinds are 'json', 'cbor', 'msgpack', 'binary'
                If None then its extracted from sad or uses default .Kind
            ilk (str | None): desired ilk packet type str value of Ilks
                If None then its extracted from sad or uses default .Ilk
            saids (dict): of keyed by label of codes for saidive fields to
                override defaults given in .Fields for a given ilk.
                If None then use defaults
                Code assignment for each saidive field in desending priority:
                   - the code provided in saids when not None
                   - the code extracted from sad[said label] when valid CESR
                   - the code provided in .Fields...saids
        """
        sproto = svrsn = skind = silk = None
        if sad and 'v' in sad:  # attempt to get from vs in sad
            try:  # extract version string elements as defaults if provided
                sproto, svrsn, skind, _, _ = deversify(sad["v"])
            except VersionError as ex:
                pass
            else:
                silk = sad.get('t')  # if 't' not in sad .get returns None which may be valid

        if proto is None:
            proto = sproto if sproto is not None else self.Proto

        if proto not in self.Fields:
            raise SerializeError(f"Invalid protocol={proto}.")

        if self.Protocol and proto != self.Protocol:  # required by class
            raise SerializeError(f"Required protocol={self.Protocol}, got "
                                 f"protocol={proto} instead.")

        if self.genus not in GenDex:  # ensures self.genus != None
            raise SerializeError(f"Invalid genus={self.genus}.")

        if getattr(GenDex, proto, None) != self.genus:
            raise SerializeError(f"Incompatible protocol={proto} with "
                                 f"genus={self.genus}.")

        if vrsn is None:
            vrsn = svrsn if svrsn is not None else self.Vrsn

        if vrsn not in self.Fields[proto]:
            raise SerializeError(f"Invalid version={vrsn} for protocol={proto}.")

        if vrsn.major > self.gvrsn.major:
            raise SerializeError(f"Incompatible major protocol version={vrsn} "
                                 f"with major genus version={self.gvrsn}.")

        if kind is None:
            kind = skind if skind is not None else self.Kind

        if ilk is None:  # default is first ilk in Fields for given proto vrsn
            ilk = (silk if silk is not None else
                   list(self.Fields[proto][vrsn])[0])  # list(dict) gives list of keys

        if kind not in Kinds:
            raise SerializeError(f"Invalid serialization kind = {kind}")

        if ilk not in self.Fields[proto][vrsn]:
            raise SerializeError(f"Invalid packet type (ilk) = {ilk} for"
                                  f"protocol = {proto}.")

        fields = self.Fields[proto][vrsn][ilk]  # get FieldDom of fields

        alls = fields.alls  # faster local reference
        oalls = oset(alls)  # ordereset of field labels
        oopts = oset(fields.opts)  # ordereset of field labels
        oreqs = oalls - oopts  # required fields


        if not sad:  # empty or None so create sad dict
            sad = {}

        # ensure all required fields are in sad. If not provide default
        for label in oreqs:
            if label not in sad:
                value = alls[label]
                if helping.nonStringIterable(value):
                    value = copy.copy(value)  # copy iterable defaults
                sad[label] = value

        sadold = sad
        sad = {}
        for label in oalls:  # make sure all fields are in correct order
            if label in sadold:
                sad[label] = sadold[label]

        for label in sadold:  # copy extras if any
            if label not in sad:
                sad[label] = sadold[label]

        if 't' in sad:  # when packet type field then force ilk
            sad['t'] = ilk  # assign ilk

        # ensure required fields are present and all fields are ordered wrt alls
        oskeys = oset(sad)  # ordered set of keys in sad (skeys)
        osexts = oskeys - oalls  # get ordered set of extras in sad (sexts)
        if osexts and fields.strict:
            raise SerializeError(f"Unallowed extra field(s) = {list(osexts)} "
                                 f"in sad.")

        osopts = oskeys - oreqs - osexts  # subset of opts in sad

        osalls = oalls - (oopts - osopts)  # subset of alls without missing opts in sad

        for k, v in fields.alts.items():
            if k in osopts and v in osopts:
                raise SerializeError(f"Unallowed, alternate fields '{k}' "
                                          f"and '{v}' both present in sad.")

        # can't do set math osalls == oskeys - osexts becasue of osexts might be
        # out-of-order so have to iterate to ensure osexts if any appear in oskeys
        # after all fields in osalls
        for i, label in enumerate(osalls):
            if oskeys[i] != label:
                raise SerializeError(f"Missing or out-of-order field = {label} "
                                            f"from = {list(osalls)} in sad.")

        # said field labels are not order dependent with respect to all fields
        # in sad so use set() to test inclusion
        _saids = copy.copy(fields.saids)  # get copy of defaults
        if not (set(_saids) <= set(alls)):
            raise SerializeError(f"Missing one or more required said fields "
                                 f"from {list(_saids)} in sad = {sad}.")

        # override saidive defaults
        for label in _saids:
            if saids and label in saids:  # use parameter override
                _saids[label] = saids[label]
            else:
                try:  # use sad field override
                    _saids[label] = Matter(qb64=sad[label]).code
                except Exception:
                    pass  # no override

            if _saids[label] in DigDex:  # if digestive then fill with dummy
                sad[label] = self.Dummy * Matter.Sizes[_saids[label]].fs


        if 'v' not in sad:  # ensures that 'v' is always required by .Labels
            raise SerializeError(f"Missing requires version string field 'v'"
                                          f" in sad = {sad}.")

        if kind in (Kinds.json, Kinds.cbor, Kinds.mgpk):
            # this size of sad needs to be computed based on actual version string span
            # since not same for all versions
            sad['v'] = self.Dummy * self.Spans[vrsn]  # ensure span of vs is dummied MAXVERFULLSPAN

            raw = self.dumps(sad, kind)  # get size of sad with fully dummied vs and saids
            size = len(raw)

            # generate new version string with correct size
            vs = versify(protocol=proto, version=vrsn, kind=kind, size=size)
            sad["v"] = vs  # update version string in sad
            # now have correctly sized version string in sad


        # compute saidive digestive field values using raw from sized dummied sad
        raw = self.dumps(sad, kind=kind, proto=proto, vrsn=vrsn)  # serialize sized dummied sad
        for label, code in _saids.items():
            if code in DigDex:  # subclass override if non digestive allowed
                sad[label] = Diger(ser=raw, code=code).qb64

        raw = self.dumps(sad, kind=kind, proto=proto, vrsn=vrsn)  # compute final raw
        if kind == Kinds.cesr:# cesr kind version string does not set size
            size = len(raw) # size of whole message

        self._raw = raw
        self._sad = sad
        self._proto = proto
        self._vrsn = vrsn
        self._kind = kind
        self._size = size


    def _inhale(self, raw, *, smellage=None):
        """Deserializes raw.
        Parses serilized event ser of serialization kind and assigns to
        instance attributes and returns tuple of associated elements.

        As classmethod enables testing parsing raw serder values. This can be
        called on self as well because it only ever accesses clas attributes
        not instance attributes.

        Returns: tuple (sad, proto, vrsn, kind, size) where:
            sad (dict): serializable attribute dict of saidified data
            proto (str): value of Protocols (Protocolage) protocol type
            vrsn (Versionage | None): tuple of (major, minor) version ints
                None means do not enforce version
            kind (str): value of Serials (Serialage) serialization kind

        Parameters:
            clas (Serder): class reference
            raw (bytes): serialized sad message
            smellage (Smellage | None): instance of deconstructed version string
                elements. If none or empty ignore otherwise assume that raw
                already had its version string extracted (reaped) into the
                elements of smellage.



        """
        if smellage:  # passed in so don't need to smell raw again
            proto, vrsn, kind, size, gvrsn = smellage  # tuple unpack
        else:  # not passed in so smell raw
            proto, vrsn, kind, size, gvrsn = smell(raw)

        sad = self.loads(raw=raw, size=size, kind=kind)
        # ._gvrsn may be set in loads when CESR native deserialization provides _gvrsn

        if "v" not in sad:  # Regex does not check for version string label itself
            raise FieldError(f"Missing version string field in {sad}.")

        # cypto opts want bytes not bytearray
        self._raw = bytes(raw[:size])  # make copy so strip not affect
        self._sad = sad
        self._proto = proto
        self._vrsn = vrsn
        self._kind = kind
        self._size = size



    def loads(self, raw, size=None, kind=Kinds.json):
        """method to handle deserialization by kind
        assumes already sniffed and smelled to determine
        serialization size and kind

        Returns:
           sad (dict | list): deserialized dict or list. Assumes attribute
                dict of saidified data.

        Parameters:
           raw (bytes | bytearray): raw serialization to deserialze as dict
           size (int): number of bytes to consume for the deserialization.
                       If None then consume all bytes in raw
           kind (str): value of Serials (Serialage) serialization kind
                       "JSON", "MGPK", "CBOR"

        Notes:
            loads of json uses str whereas loads of cbor and msgpack use bytes
        """
        if kind == Kinds.json:
            try:
                sad = json.loads(raw[:size].decode("utf-8"))
            except Exception as ex:
                raise DeserializeError(f"Error deserializing JSON: "
                    f"{raw[:size].decode('utf-8')}") from ex

        elif kind == Kinds.mgpk:
            try:
                sad = msgpack.loads(raw[:size])
            except Exception as ex:
                raise DeserializeError(f"Error deserializing MGPK: "
                    f"{raw[:size].decode('utf-8')}") from ex

        elif kind == Kinds.cbor:
            try:
                sad = cbor.loads(raw[:size])
            except Exception as ex:
                raise DeserializeError(f"Error deserializing CBOR: "
                    f"{raw[:size].decode('utf-8')}") from ex

        else:
            raise DeserializeError(f"Invalid deserialization kind: {kind}")

        return sad


    def _loads(self, raw, size=None):
        """CESR native desserialization of raw

        Returns:
           sad (dict): deserialized dict of CESR native serialization.

        Parameters:
           clas (Serder): class reference
           raw (bytes |bytearray): raw serialization to deserialze as dict
           size (int): number of bytes to consume for the deserialization.
                       If None then consume all bytes in raw
        """
        # ._gvrsn may be set in loads when CESR native deserialization provides _gvrsn
        pass


    def _exhale(self, sad):
        """Serializes sad and assigns attributes.
        Asssumes all field values in sad are valid.
        Call .verify to otherwise

        Parameters:
            sad (dict): serializable attribute dict of saidified data
        """
        if "v" not in sad:
            raise SerializeError(f"Missing version string field in {sad}.")

        # extract elements so can replace size element but keep others
        proto, vrsn, kind, size, opt = deversify(sad["v"])

        raw = self.dumps(sad, kind)

        if kind in (Kinds.cesr):  # cesr kind version string does not set size
            size = len(raw) # size of whole message

        # must call .verify to ensure these are compatible
        self._raw = raw  # crypto opts want bytes not bytearray
        self._sad = sad
        self._proto = proto
        self._vrsn = vrsn
        self._kind = kind
        self._size = size


    def dumps(self, sad=None, kind=Kinds.json, proto=None, vrsn=None):
        """Method to handle serialization by kind
        Assumes sad fields are properly filled out for serialization kind.

        Returns:
            raw (bytes): serialization of sad dict using serialization kind

        Parameters:
            sad (dict | list | None)): serializable dict or list to serialize
            kind (str): value of Serials (Serialage) serialization kind
                "JSON", "MGPK", "CBOR", "CSER"
            proto (str | None): desired protocol type str value of Protocols
                If None then eventually use self.proto
            vrsn (Versionage | None): instance desired protocol version
                If None then eventually self.vrsn


        Notes:
            dumps of json uses str whereas dumps of cbor and msgpack use bytes
            crypto opts want bytes not bytearray
        """
        sad = sad if sad is not None else self.sad

        if kind == Kinds.json:
            raw = json.dumps(sad, separators=(",", ":"),
                             ensure_ascii=False).encode("utf-8")

        elif kind == Kinds.mgpk:
            raw = msgpack.dumps(sad)

        elif kind == Kinds.cbor:
            raw = cbor.dumps(sad)

        elif kind == Kinds.cesr:  # does not support list only dict
            raw = self._dumps(sad, proto=proto, vrsn=vrsn)

        else:
            raise SerializeError(f"Invalid serialization kind = {kind}")

        return raw


    def _dumps(self, sad=None, proto=None, vrsn=None):
        """CESR native serialization of sad

        Returns:
            raw (bytes): CESR native serialization of sad dict

        Parameters:
            sad (dict | None)): serializable dict to serialize
            proto (str | None): desired protocol type str value of Protocols
                If None then self.proto
            vrsn (Versionage | None): instance desired protocol version
                If None then self.vrsn

        Versioning:
            CESR native serialization includes in its fixed version field
            a version primitive that includes message protocol+protocol version
            +genus version: 0NPPPPMmmMmm (12 B64 characters)

            This assumes that genus is compatible with message
            protocol so genus is not needed. This protects from malleability attack
            and ensure compatible cesr codes especially count (group) codes.
            Primitive codes are less problematic since so far all primitive codes
            tables are backwards compatible across major versions.

        """
        sad = sad if sad is not None else self.sad
        proto = proto if proto is not None else self.proto
        vrsn = vrsn if vrsn is not None else self.vrsn

        if (self.gvrsn.major < Vrsn_2_0.major or vrsn.major < Vrsn_2_0.major):
            raise SerializeError(f"Invalid major genus version={self.gvrsn}"
                                f"or Invalid major protocol version={vrsn}"
                                f" for native CESR serialization.")

        if self.genus not in GenDex:  # ensures self.genus != None
            raise SerializeError(f"Invalid genus={self.genus}.")

        if getattr(GenDex, proto, None) != self.genus:
            raise SerializeError(f"Incompatible protocol={proto} with "
                                 f"genus={self.genus}.")




        raw = bytearray()  # message as qb64
        bdy = bytearray()  # message body as qb64
        ilks = self.Fields[proto][vrsn]  # get fields keyed by ilk

        ilk = sad.get('t')  # returns None if missing message type (ilk)
        if ilk not in ilks:  #
            raise SerializeError(f"Missing message type field "
                                 f"'t' for protocol={proto} "
                                 f"version={vrsn} with {sad=}.")

        fields = ilks[ilk]  # FieldDom for given protocol and ilk

        if fields.opts or not fields.strict:  # optional or extra fields allowed
            fixed = False  # so must use field map
        else:
            fixed = True  #fixed field


        # assumes that sad's field ordering and field inclusion is correct
        # so can serialize in order to compute saidive fields
        # need to fix ._verify and .makify to account for CESR native serialization

        if proto == Protocols.keri:
            if not fixed:  # prepend label
                pass  # raise error

            for l, v in sad.items():  # assumes valid field order & presence
                match l:  # label
                    case "v":  # protocol+version  do not use version string itself
                        val = Verser(proto=proto, vrsn=vrsn).qb64b

                    case "t":  # message type (ilk), already got ilk
                        val = Ilker(ilk=v).qb64b  # assumes same

                    case "d" | "i" | "p" | "di":  # said or aid
                        val = v.encode("utf-8")  # already primitive qb64 make qb6b

                    case "s" | "bt":  # sequence number or numeric threshold
                        val = coring.Number(numh=v).qb64b  # convert hex str

                    case "kt" | "nt": # current or next signing threshold
                        val = coring.Tholder(sith=v).limen  # convert sith str

                    case "k" | "n" | "b" | "ba" | "br":  # list of primitives
                        frame = bytearray()
                        for e in v:  # list
                            frame.extend(e.encode("utf-8"))

                        val = bytearray(Counter(Codens.GenericListGroup,
                                                count=len(frame) // 4).qb64b)
                        val.extend(frame)

                    case "c":  # list of config traits strings
                        frame = bytearray()
                        for e in v:  # list
                            frame.extend(Traitor(trait=e).qb64b)

                        val = bytearray(Counter(Codens.GenericListGroup,
                                                count=len(frame) // 4).qb64b)
                        val.extend(frame)

                    case "a":  # list of seals or field map of attributes
                        frame = bytearray()  # whole list
                        gcode = None  # code for counter for consecutive same type seals
                        gframe = bytearray()  # consecutive same type seals
                        for e in v:  # list of seal dicts
                            # need support for grouping consecutive seals of same type with same counter

                            try:
                                sealer = Sealer(crew=e)
                                code = self.ClanCodes[sealer.name]
                                if gcode and gcode == code:
                                    gframe.extend(sealer.qb64b)
                                else:
                                    if gframe:  # not same so close off and rotate group
                                        counter = Counter(code=gcode, count=len(gframe) // 4)
                                        frame.extend(counter.qb64b + gframe)
                                        gframe = bytearray()  # new group
                                    gcode = code  # new group or keep same group
                                    gframe.extend(sealer.qb64b)  # extend in new group

                            except kering.InvalidValueError:
                                if gframe:
                                    counter = Counter(code=gcode, count=len(gframe) // 4)
                                    frame.extend(counter.qb64b + gframe)
                                    gframe = bytearray()
                                    gcode = None

                                #unknown seal type so serialize as field map
                                #generic seal no count type (v, Mapping):
                                #for l, e in v.items():
                                    #pass
                                #val = bytearray(Counter(tag=""GenericMapGroup"",
                                               # count=len(frame) // 4).qb64b)
                                #val.extend(mapframe)

                        if gframe:  # close off last group if any
                            counter = Counter(code=gcode, count=len(gframe) // 4)
                            frame.extend(counter.qb64b + gframe)
                            gframe = bytearray()
                            gcode = None

                        val = bytearray(Counter(Codens.GenericListGroup,
                                                count=len(frame) // 4).qb64b)
                        val.extend(frame)

                    case _:  # if extra fields this is where logic would be
                        raise SerializeError(f"Unsupported protocol field label"
                                             f"='{l}' for protocol={proto}"
                                             f" version={vrsn}.")

                bdy.extend(val)


        elif proto == Protocols.acdc:
            for l, val in sad.items():  # assumes valid field order & presence
                if not fixed:
                    pass  # prepend label



        else:
            raise SerializeError(f"Unsupported protocol={self.proto}.")


        # prepend count code for message
        if fixed:

            raw = bytearray(Counter(Codens.FixedMessageBodyGroup,
                                    count=len(bdy) // 4).qb64b)
            raw.extend(bdy)
        else:
            pass


        return raw



    def compare(self, said=None):
        """Utility method to allow comparison of own .said digest of .raw
        with some other purported said of .raw

        Returns:
            success (bool): True if said matches self.saidb  via string
               equality. Converts said to bytes if unicode


        Parameters:
            said (bytes | str): qb64b or qb64 digest to compare with .said
        """
        if said is not None:
            if hasattr(said, "encode"):
                said = said.encode('utf-8')  # makes bytes

            return said == self.saidb  # str match bool

        else:
            raise ValidationError(f"Uncomparable saids.")



    def pretty(self, *, size=None):
        """Utility method to pretty print .sad as JSON str.
        Returns:
            pretty (str):  JSON of .sad with pretty formatting

        Pararmeters:
            size (int | None): size limit. None means not limit.
                Enables protection against syslog error when
                exceeding UDP MTU (max trans unit) for syslog applications.
                Guaranteed IPv4 MTU is 576, and IPv6 MTU is 1280.
                Most broadband routers have an UDP MTU set to 1454.
                Must include not just payload but UDP/IP header in
                MTU calculation. So must leave room for either UDP/IpV4 or
                the bigger UDP/IPv6 header.
                Except for old IoT hardware, modern implementations all
                support IPv6 so 1024 is usually a safe value for payload.
        """
        return json.dumps(self._sad, indent=1)[:size]


    @property
    def raw(self):
        """raw property getter
        Returns:
            raw (bytes):  serialized version
        """
        return self._raw


    @property
    def sad(self):
        """sad property getter
        Returns:
            sad (dict): serializable attribute dict (saidified data)
        """
        return dict(self._sad)  # return copy


    @property
    def genus(self):
        """genus (CESR genu code) property getter

        Returns:
            genus (stre): CESR genus code for this Serder
        """
        return self._genus


    @property
    def gvrsn(self):
        """gvrsn (CESR genus version) property getter

        Returns:
            gvrsn (Versionage): instance, CESR genus code table version for this Serder
        """
        return self._gvrsn


    @property
    def kind(self):
        """kind property getter
        Returns:
            kind (str): value of Serials (Serialage)"""
        return self._kind


    @property
    def proto(self):
        """proto property getter,
        protocol identifier type value of Protocolage such as 'KERI' or 'ACDC'

        Returns:
            proto (str): Protocolage value as protocol type
        """
        return self._proto

    @property
    def protocol(self):
        """protocp; property getter, alias of .proto
        protocol identifier type value of Protocolage such as 'KERI' or 'ACDC'

        Returns:
            protocol (str): Protocolage value as protocol type
        """
        return self.proto


    @property
    def vrsn(self):
        """vrsn (version) property getter

        Returns:
            vrsn (Versionage): instance of protocol version for this Serder
        """
        return self._vrsn

    @property
    def version(self):
        """version property getter, alias of .vrsn

        Returns:
            version (Versionage): instance of protocol version for this Serder
        """
        return self.vrsn


    @property
    def size(self):
        """size property getter
        Returns:
            size (int): number of bytes in .raw
        """
        return self._size


    @property
    def said(self):
        """said property getter
        Returns:
           said (str): qb64
        """
        if not self.Fields[self.proto][self.vrsn][self.ilk].saids and 'd' in self._sad:
            return self._sad['d']  # special case for non-saidive messages like rct
        return self._said


    @property
    def saidb(self):
        """saidb property getter
        Returns:
            saidb (bytes): qb64b of said  of .saider
        """
        return self.said.encode("utf-8") if self.said is not None else None


    @property
    def ilk(self):
        """ilk property getter
        Returns:
            ilk (str): pracket type given by sad['t'] if any
        """
        return self._sad.get('t')  # returns None if 't' not in sad



class SerderKERI(Serder):
    """SerderKERI is Serder subclass with Labels for KERI packet types (ilks) and
       properties for exposing field values of KERI messages

       See docs for Serder
    """
    #override in subclass to enforce specific protocol
    Protocol = Protocols.keri  # required protocol, None means any in Protocols is ok
    Proto = Protocols.keri  # default protocol type


    def _verify(self, **kwa):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        super(SerderKERI, self)._verify(**kwa)

        allkeys = list(self.Fields[self.proto][self.vrsn][self.ilk].alls)
        keys = list(self.sad)
        if allkeys != keys:
            raise ValidationError(f"Invalid top level field list. Expected "
                                  f"{allkeys} got {keys}.")

        if (self.vrsn.major < 2 and self.vrsn.minor < 1 and
            self.ilk in (Ilks.qry, Ilks.rpy, Ilks.pro, Ilks.bar, Ilks.exn)):
                pass  # non prefixive ilks do not have 'i' field
        else:  # verify pre 'i' field
            try:
                code = Matter(qb64=self.pre).code
            except Exception as ex:
                raise ValidationError(f"Invalid identifier prefix = "
                                      f"{self.pre}.") from ex

            if self.ilk in (Ilks.dip, Ilks.drt):
                idex = DigDex  # delegatee must be digestive prefix
            else:
                idex = PreDex  # non delegatee may be non digest

            if code not in idex:
                raise ValidationError(f"Invalid identifier prefix code = {code}.")


            if self.ilk in (Ilks.icp, Ilks.dip, Ilks.rot, Ilks.drt):  # est event
                if self.ilk in (Ilks.icp, Ilks.dip):  # inceptive event
                    if code in PreNonDigDex:
                        if len(self.keys) != 1:
                            raise ValidationError(f"Invalid keys = {self.keys} "
                                                  "for non-digestive prefix "
                                                  f"{code=}.")

                        if self.tholder.sith != '1':
                            raise ValidationError(f"Invalid signing threshold ="
                                                  f" {self.tholder.sith} for "
                                                  f"non-digestive prefix {code=}.")

                        if self.pre != self.keys[0]:
                            raise ValidationError(f"Mismatch prefix = {self.pre} and"
                                                  f" zeroth key = {self.keys[0]} for "
                                                  f" non-digestive prefix {code=}.")

                # non-transferable pre validations
                if code in NonTransDex:
                    if self.ndigs:  # when field missing returns None
                        raise ValidationError(f"Non-transferable code = {code} with"
                                              f" non-empty nxt = {self.ndigs}.")

                    if self.backs:  # when field missing returns None
                        raise ValidationError("Non-transferable code = {code} with"
                                              f" non-empty backers = {self.backs}.")

                    if self.seals:  # when field missing returns None
                        raise ValidationError("Non-transferable code = {code} with"
                                              f" non-empty seals = {self.seals}.")

        if self.ilk in (Ilks.dip):  # validate delpre
            try:
                code = Matter(qb64=self.delpre).code
            except Exception as ex:
                raise ValidationError(f"Invalid delegator prefix = "
                                      f"{self.delpre}.") from ex

            if code not in PreDex:  # delegator must be valid prefix code
                raise ValidationError(f"Invalid delegator prefix code = {code}.")


    @property
    def estive(self):  # establishative
        """ Returns True if Serder represents an establishment event """
        return (self._sad["t"] in (Ilks.icp, Ilks.rot, Ilks.dip, Ilks.drt)
                     if "t" in self._sad else False)


    @property
    def ked(self):
        """
        Returns:
            ked (dict): key event dict property getter. Alias for .sad
        """
        return self.sad


    @property
    def pre(self):
        """
        Returns:
           pre (str): qb64  of .sad["i"] identifier prefix property getter
        """
        return self._sad.get("i")


    @property
    def preb(self):
        """
        Returns:
        preb (bytes): qb64b  of .pre identifier prefix property getter as bytes
        """
        return self.pre.encode("utf-8") if self.pre is not None else None


    @property
    def sner(self):
        """Number instance of sequence number, sner property getter

        Returns:
            (Number): of ._sad["s"] hex number str converted
        """
        # auto converts hex num str to int
        return Number(num=self._sad["s"]) if 's' in self._sad else None


    @property
    def sn(self):
        """Sequence number, sn property getter
        Returns:
            sn (int): of .sner.num from .sad["s"]
        """
        return self.sner.num if self.sner is not None else None


    @property
    def snh(self):
        """Sequence number hex str, snh property getter
        Returns:
            snh (hex str): of .sner.numh from .sad["s"]
        """
        return self.sner.numh if self.sner is not None else None


    @property
    def seals(self):
        """Seals property getter

        Returns:
            seals (list): from ._sad["a"]
        """
        return self._sad.get("a")

    #Properties of inceptive Serders ilks in (icp, dip) and version2 estive serders

    @property
    def traits(self):
        """Traits list property getter  (config traits)

        Returns:
            traits (list): from ._sad["c"]
        """
        return self._sad.get("c")


    #Properties of estive Serders ilks in  (icp, rot, dip, drt)
    @property
    def tholder(self):
        """Tholder property getter

        Returns:
            tholder (Tholder): instance as converted from ._sad['kt']
                or None if missing.

        """
        return Tholder(sith=self._sad["kt"]) if "kt" in self._sad else None


    @property
    def keys(self):
        """Returns list of qb64 keys from ._sad['k'].
        One for each key.
        keys property getter
        """
        return self._sad.get("k")


    @property
    def verfers(self):
        """Returns list of Verfer instances as converted from ._sad['k'].
        One for each key.
        verfers property getter
        """
        keys = self._sad.get("k")
        return [Verfer(qb64=key) for key in keys] if keys is not None else None


    @property
    def ntholder(self):
        """Returns Tholder instance as converted from ._sad['nt'] or None if missing.

        """
        return Tholder(sith=self._sad["nt"]) if "nt" in self._sad else None


    @property
    def ndigs(self):
        """
        Returns:
            (list): digs
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return None

        return self._sad.get("n")

    @property
    def ndigers(self):
        """NDigers property getter

        Returns:
            ndigers (list[Diger]): instance as converted from ._sad['n'].
            One for each next key digests.
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return None

        digs = self._sad.get("n")
        return [Diger(qb64=dig) for dig in digs] if digs is not None else None


    @property
    def bner(self):  # toader
        """
        bner (Number of backer TOAD threshold of accountable duplicity property getter
        Returns:
            (Number): of ._sad["bt"] hex number str converted. Auto converts
            hex num str to int
        """
        return Number(num=self._sad["bt"]) if 'bt' in self._sad else None


    @property
    def bn(self):
        """
        bn (backer TOAD number) property getter
        Returns:
            bn (int): of .bner.num from .ked["bt"]
        """
        return self.bner.num if self.bner is not None else None


    # properties for incentive Serders like icp, dip
    @property
    def backs(self):
        """Backers property getter

        Returns:
            backs (list[str]): aids qb64 from ._sad['b'].
                           One for each backer (witness).

        """
        return self._sad.get("b")


    @property
    def berfers(self):
        """Berfers property getter
        Returns list of Verfer instances as converted from ._sad['b'].
                One for each backer (witness).

        """
        baks = self._sad.get("b")
        return [Verfer(qb64=bak) for bak in baks] if baks is not None else None


    # properties for priorative Serders like ixn rot drt

    @property
    def prior(self):
        """Prior property getter
        Returns:
            prior (str): said qb64 of prior event from ._sad['p'].

        """
        return self._sad.get("p")


    @property
    def priorb(self):
        """Priorb bytes property getter
        Returns:
            priorb (str): said qb64b of prior event from ._sad['p'].

        """
        return self.prior.encode("utf-8") if self.prior is not None else None


    # properties for rotative Serders like rot drt

    @property
    def cuts(self):
        """Cuts property getter
        Returns list of aids of instances as converted from ._sad['br'].
                 One for each backer (witness) to be cut (removed).

        """
        return self._sad.get("br")


    @property
    def adds(self):
        """Adds property getter
        Returns list of aids of instances as converted from ._sad['ba'].
                 One for each backer (witness) to be added.

        """
        return self._sad.get("ba")


    #Properties for delegated Serders ilks in (dip, drt)

    @property
    def delpre(self):
        """
        Returns:
           delpre (str): qb64  of .sad["di"] delegator ID prefix property getter
        """
        return self._sad.get("di")


    @property
    def delpreb(self):
        """
        Returns:
        delpreb (bytes): qb64b  of .delpre property getter as bytes
        """
        return self.delpre.encode("utf-8") if self.delpre is not None else None

    #Propertives for dated Serders, qry, rpy, pro, bar, exn

    @property
    def stamp(self):
        """
        Returns:
           stamp (str): date-time-stamp sad["dt"]. RFC-3339 profile of ISO-8601
                datetime of creation of message or data
        """
        return self._sad.get("dt")


    #Properties for exn  exchange


    #Properties for vcp  (registry  inception event)
    @property
    def uuid(self):
        """uuid property getter

        Returns:
           uuid (str): qb64  of .sad["u"] salty nonce
        """
        return self._sad.get("u")

    @property
    def nonce(self):
        """
        should be deprecated

        Returns:
           nonce (str): alias for .uuid property
        """
        if self.vrsn.major < 2 and self.vrsn.minor < 1 and self.ilk == Ilks.vcp:
            return self._sad.get("n")
        else:
            return self.uuid



class SerderACDC(Serder):
    """SerderACDC is Serder subclass with Labels for ACDC packet types (ilks) and
       properties for exposing field values of ACDC messages

       See docs for Serder
    """
    #override in subclass to enforce specific protocol
    Protocol = Protocols.acdc  # required protocol, None means any in Protocols is ok
    Proto = Protocols.acdc  # default protocol type



    def _verify(self, **kwa):
        """Verifies said(s) in sad against raw
        Override for protocol and ilk specific verification behavior. Especially
        for inceptive ilks that have more than one said field like a said derived
        identifier prefix.

        Raises a ValidationError (or subclass) if any verification fails

        """
        super(SerderACDC, self)._verify(**kwa)

        try:
            code = Matter(qb64=self.issuer).code
        except Exception as ex:
            raise ValidationError(f"Invalid issuer AID = "
                                  f"{self.issuer}.") from ex

        if code not in PreDex:
            raise ValidationError(f"Invalid issuer AID code = {code}.")

    @property
    def uuid(self):
        """uuid property getter
        Optional fields return None when not present
        Returns:
           uuid (str | None): qb64  of .sad["u"] salty nonce
        """
        return self._sad.get("u")


    @property
    def uuidb(self):
        """uuid property getter (uuid bytes)
        Optional fields return None when not present
        Returns:
           uuidb (bytes | None): qb64b  of .sad["u"] salty nonce as bytes
        """
        return self.uuid.encode("utf-8") if self.uuid is not None else None


    @property
    def issuer(self):
        """issuer property getter (issuer AID)
        Optional fields return None when not present
        Returns:
           issuer (str | None): qb64  of .sad["i"] issuer AID
        """
        return self._sad.get('i')


    @property
    def issuerb(self):
        """issuerb property getter (issuer AID bytes)
        Optional fields return None when not present
        Returns:
        issuerb (bytes | None): qb64b  of .issuer AID as bytes
        """
        return self.issuer.encode("utf-8") if self.issuer is not None else None


    @property
    def regi(self):
        """regi property getter (registry identifier SAID)
        Optional fields return None when not present
        Returns:
           regi (str | None): qb64  of .sad["ri"] registry SAID
        """
        return self._sad.get('ri')


    @property
    def regib(self):
        """regib property getter (registry identifier SAID bytes)
        Optional fields return None when not present
        Returns:
        regib (bytes | None): qb64b  of .issuer AID as bytes
        """
        return self.issuer.encode("utf-8") if self.issuer is not None else None


    @property
    def schema(self):
        """schema block or SAID property getter
        Optional fields return None when not present
        Returns:
            schema (dict | str | None): from ._sad["s"]
        """
        return self._sad.get('s')


    @property
    def attrib(self):
        """attrib block or SAID property getter (attribute)
        Optional fields return None when not present
        Returns:
            attrib (dict | str | None): from ._sad["a"]
        """
        return self._sad.get("a")


    @property
    def issuee(self):
        """ise property getter (issuee AID)
        Optional fields return None when not present
        Returns:
           issuee (str | None): qb64  of .sad["a"]["i"] issuee AID
        """
        try:
            return self.attrib.get['i']
        except:
            return None


    @property
    def issueeb(self):
        """isrb property getter (issuee AID bytes)
        Optional fields return None when not present
        Returns:
        issueeb (bytes | None): qb64b  of .issuee AID as bytes
        """
        return self.issuee.encode("utf-8") if self.issuee is not None else None


    @property
    def attagg(self):
        """Attagg block property getter (attribute aggregate)
        Optional fields return None when not present
        Returns:
            attagg (dict | str): from ._sad["A"]
        """
        return self._sad.get("A")


    @property
    def edge(self):
        """Edge block property getter
        Optional fields return None when not present
        Returns:
            edge (dict | str): from ._sad["e"]
        """
        return self._sad.get("e")


    @property
    def rule(self):
        """Rule block property getter
        Optional fields return None when not present

        Returns:
            rule (dict | str): from ._sad["r"]
        """
        return self._sad.get("r") # or {}  # need to fix logic so can remove or since optional

    # ToDo Schemer property getter. Schemer object should change name to Schemar



# -*- encoding: utf-8 -*-
"""
keri.app.cli.commands module

"""
import argparse
import secrets

import string
from hio.base import doing

parser = argparse.ArgumentParser(description='Print a new random salt')
parser.set_defaults(handler=lambda args: handler(args))

PASSCODE_SIZE = 21
PASSCODE_CHARS = string.ascii_lowercase + string.ascii_uppercase + '123456789'


def handler(_):
    return [doing.doify(salt)]


def salt(tymth, tock=0.0):
    """ Command line version handler
    """
    _ = (yield tock)

    code = []
    for x in range(PASSCODE_SIZE):
        code.append(PASSCODE_CHARS[secrets.randbelow(len(PASSCODE_CHARS))])

    print("".join(code))




# -*- coding: utf-8 -*-
"""
keri.core.indexing module

Provides versioning support for Indexer classes and codes
"""
from collections import namedtuple, deque
from dataclasses import dataclass, astuple, asdict
from base64 import urlsafe_b64encode as encodeB64
from base64 import urlsafe_b64decode as decodeB64

import pysodium


from ..kering import (EmptyMaterialError, RawMaterialError, SoftMaterialError,
                      InvalidCodeError, InvalidSoftError,
                      InvalidSizeError,
                      InvalidCodeSizeError, InvalidVarIndexError,
                      InvalidVarSizeError, InvalidVarRawSizeError,
                      ConversionError, InvalidValueError, InvalidTypeError,
                      ValidationError, VersionError, DerivationError,
                      EmptyListError,
                      ShortageError, UnexpectedCodeError, DeserializeError,
                      UnexpectedCountCodeError, UnexpectedOpCodeError)

from ..help import helping
from ..help.helping import (sceil, intToB64, b64ToInt,
                            codeB64ToB2, codeB2ToB64, nabSextets)


@dataclass(frozen=True)
class IndexerCodex:
    """ IndexerCodex is codex hard (stable) part of all indexer derivation codes.

    Codes indicate which list of keys, current and/or prior next, index is for:

        _Sig:           Indices in code may appear in both current signing and
                        prior next key lists when event has both current and prior
                        next key lists. Two character code table has only one index
                        so must be the same for both lists. Other index if for
                        prior next.
                        The indices may be different in those code tables which
                        have two sets of indices.

        _Crt_Sig:       Index in code for current signing key list only.

        _Big_:          Big index values


    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """

    Ed25519_Sig: str = 'A'  # Ed25519 sig appears same in both lists if any.
    Ed25519_Crt_Sig: str = 'B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Sig: str = 'C'  # ECDSA secp256k1 sig appears same in both lists if any.
    ECDSA_256k1_Crt_Sig: str = 'D'  # ECDSA secp256k1 sig appears in current list.
    ECDSA_256r1_Sig: str = "E"  # ECDSA secp256r1 sig appears same in both lists if any.
    ECDSA_256r1_Crt_Sig: str = "F"  # ECDSA secp256r1 sig appears in current list.
    Ed448_Sig: str = '0A'  # Ed448 signature appears in both lists.
    Ed448_Crt_Sig: str = '0B'  # Ed448 signature appears in current list only.
    Ed25519_Big_Sig: str = '2A'  # Ed25519 sig appears in both lists.
    Ed25519_Big_Crt_Sig: str = '2B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Big_Sig: str = '2C'  # ECDSA secp256k1 sig appears in both lists.
    ECDSA_256k1_Big_Crt_Sig: str = '2D'  # ECDSA secp256k1 sig appears in current list only.
    ECDSA_256r1_Big_Sig: str = "2E"  # ECDSA secp256r1 sig appears in both lists.
    ECDSA_256r1_Big_Crt_Sig: str = "2F"  # ECDSA secp256r1 sig appears in current list only.
    Ed448_Big_Sig: str = '3A'  # Ed448 signature appears in both lists.
    Ed448_Big_Crt_Sig: str = '3B'  # Ed448 signature appears in current list only.
    TBD0: str = '0z'  # Test of Var len label L=N*4 <= 4095 char quadlets includes code
    TBD1: str = '1z'  # Test of index sig lead 1
    TBD4: str = '4z'  # Test of index sig lead 1 big

    def __iter__(self):
        return iter(astuple(self))  # enables inclusion test with "in"

IdrDex = IndexerCodex()


@dataclass(frozen=True)
class IndexedSigCodex:
    """IndexedSigCodex is codex all indexed signature derivation codes.

    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    Ed25519_Sig: str = 'A'  # Ed25519 sig appears same in both lists if any.
    Ed25519_Crt_Sig: str = 'B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Sig: str = 'C'  # ECDSA secp256k1 sig appears same in both lists if any.
    ECDSA_256k1_Crt_Sig: str = 'D'  # ECDSA secp256k1 sig appears in current list.
    ECDSA_256r1_Sig: str = "E"  # ECDSA secp256r1 sig appears same in both lists if any.
    ECDSA_256r1_Crt_Sig: str = "F"  # ECDSA secp256r1 sig appears in current list.
    Ed448_Sig: str = '0A'  # Ed448 signature appears in both lists.
    Ed448_Crt_Sig: str = '0B'  # Ed448 signature appears in current list only.
    Ed25519_Big_Sig: str = '2A'  # Ed25519 sig appears in both lists.
    Ed25519_Big_Crt_Sig: str = '2B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Big_Sig: str = '2C'  # ECDSA secp256k1 sig appears in both lists.
    ECDSA_256k1_Big_Crt_Sig: str = '2D'  # ECDSA secp256k1 sig appears in current list only.
    ECDSA_256r1_Big_Sig: str = "2E"  # ECDSA secp256r1 sig appears in both lists.
    ECDSA_256r1_Big_Crt_Sig: str = "2F"  # ECDSA secp256r1 sig appears in current list only.
    Ed448_Big_Sig: str = '3A'  # Ed448 signature appears in both lists.
    Ed448_Big_Crt_Sig: str = '3B'  # Ed448 signature appears in current list only.

    def __iter__(self):
        return iter(astuple(self))

IdxSigDex = IndexedSigCodex()  # Make instance


@dataclass(frozen=True)
class IndexedCurrentSigCodex:
    """IndexedCurrentSigCodex is codex indexed signature codes for current list.

    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    Ed25519_Crt_Sig: str = 'B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Crt_Sig: str = 'D'  # ECDSA secp256k1 sig appears in current list only.
    ECDSA_256r1_Crt_Sig: str = "F"  # ECDSA secp256r1 sig appears in current list.
    Ed448_Crt_Sig: str = '0B'  # Ed448 signature appears in current list only.
    Ed25519_Big_Crt_Sig: str = '2B'  # Ed25519 sig appears in current list only.
    ECDSA_256k1_Big_Crt_Sig: str = '2D'  # ECDSA secp256k1 sig appears in current list only.
    ECDSA_256r1_Big_Crt_Sig: str = "2F"  # ECDSA secp256r1 sig appears in current list only.
    Ed448_Big_Crt_Sig: str = '3B'  # Ed448 signature appears in current list only.

    def __iter__(self):
        return iter(astuple(self))

IdxCrtSigDex = IndexedCurrentSigCodex()  # Make instance



@dataclass(frozen=True)
class IndexedBothSigCodex:
    """IndexedBothSigCodex is codex indexed signature codes for both lists.

    Only provide defined codes.
    Undefined are left out so that inclusion(exclusion) via 'in' operator works.
    """
    Ed25519_Sig: str = 'A'  # Ed25519 sig appears same in both lists if any.
    ECDSA_256k1_Sig: str = 'C'  # ECDSA secp256k1 sig appears same in both lists if any.
    ECDSA_256r1_Sig: str = "E"  # ECDSA secp256r1 sig appears same in both lists if any.
    Ed448_Sig: str = '0A'  # Ed448 signature appears in both lists.
    Ed25519_Big_Sig: str = '2A'  # Ed25519 sig appears in both listsy.
    ECDSA_256k1_Big_Sig: str = '2C'  # ECDSA secp256k1 sig appears in both lists.
    ECDSA_256r1_Big_Sig: str = "2E"  # ECDSA secp256r1 sig appears in both lists.
    Ed448_Big_Sig: str = '3A'  # Ed448 signature appears in both lists.

    def __iter__(self):
        return iter(astuple(self))

IdxBthSigDex = IndexedBothSigCodex()  # Make instance

# namedtuple for size entries in Incexer derivation code tables
# hs is the hard size int number of chars in hard (stable) part of code
# ss is the soft size int number of chars in soft (unstable) part of code
# os is the other size int number of chars in other index part of soft
# ms = ss - os main index size computed
# fs is the full size int number of chars in code plus appended material if any
# ls is the lead size int number of bytes to pre-pad pre-converted raw binary
Xizage = namedtuple("Xizage", "hs ss os fs ls")

class Indexer:
    """ Indexer is fully qualified cryptographic material primitive base class for
    indexed primitives. In special cases some codes in the Index code table
    may be of variable length (i.e. not indexed) when the full size table entry
    is None. In that case the index is used instread as the length.

    Sub classes are derivation code and key event element context specific.

    Includes the following attributes and properties:

    Attributes:

    Properties:
        code is str of stable (hard) part of derivation code
        raw (bytes): unqualified crypto material usable for crypto operations
        index (int): main index offset into list or length of material
        ondex (int | None): other index offset into list or length of material
        qb64b (bytes): fully qualified Base64 crypto material
        qb64 (str | bytes):  fully qualified Base64 crypto material
        qb2 (bytes): fully qualified binary crypto material

    Hidden:
        ._code (str): value for .code property
        ._raw (bytes): value for .raw property
        ._index (int): value for .index property
        ._ondex (int): value for .ondex property
        ._infil is method to compute fully qualified Base64 from .raw and .code
        ._binfil is method to compute fully qualified Base2 from .raw and .code
        ._exfil is method to extract .code and .raw from fully qualified Base64
        ._bexfil is method to extract .code and .raw from fully qualified Base2

    """
    # Hards table maps from bytes Base64 first code char to int of hard size, hs,
    # (stable) of code. The soft size, ss, (unstable) is always > 0 for Indexer.
    Hards = ({chr(c): 1 for c in range(65, 65 + 26)})
    Hards.update({chr(c): 1 for c in range(97, 97 + 26)})
    Hards.update([('0', 2), ('1', 2), ('2', 2), ('3', 2), ('4', 2)])
    # Sizes table maps hs chars of code to Xizage namedtuple of (hs, ss, os, fs, ls)
    # where hs is hard size, ss is soft size, os is other index size,
    # and fs is full size, ls is lead size.
    # where ss includes os, so main index size ms = ss - os
    # soft size, ss, should always be  > 0 for Indexer
    Sizes = {
        'A': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        'B': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        'C': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        'D': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        'E': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        'F': Xizage(hs=1, ss=1, os=0, fs=88, ls=0),
        '0A': Xizage(hs=2, ss=2, os=1, fs=156, ls=0),
        '0B': Xizage(hs=2, ss=2, os=1, fs=156, ls=0),
        '2A': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '2B': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '2C': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '2D': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '2E': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '2F': Xizage(hs=2, ss=4, os=2, fs=92, ls=0),
        '3A': Xizage(hs=2, ss=6, os=3, fs=160, ls=0),
        '3B': Xizage(hs=2, ss=6, os=3, fs=160, ls=0),
        '0z': Xizage(hs=2, ss=2, os=0, fs=None, ls=0),
        '1z': Xizage(hs=2, ss=2, os=1, fs=76, ls=1),
        '4z': Xizage(hs=2, ss=6, os=3, fs=80, ls=1),
    }
    # Bards table maps to hard size, hs, of code from bytes holding sextets
    # converted from first code char. Used for ._bexfil.
    Bards = ({codeB64ToB2(c): hs for c, hs in Hards.items()})

    Codes = asdict(IdrDex)  # map code name to code
    Names = {val : key for key, val in Codes.items()} # invert map code to code name



    def __init__(self, raw=None, code=IdrDex.Ed25519_Sig, index=0, ondex=None,
                 qb64b=None, qb64=None, qb2=None, strip=False, **kwa):
        """
        Validate as fully qualified
        Parameters:
            raw (bytes): unqualified crypto material usable for crypto operations
            code is str of stable (hard) part of derivation code
            index (int): main index offset into list or length of material
            ondex (int | None): other index offset into list or length of material
            qb64b (bytes): fully qualified Base64 crypto material
            qb64 (str | bytes):  fully qualified Base64 crypto material
            qb2 (bytes): fully qualified binary crypto material
            strip (bool): True means strip counter contents from input stream
                bytearray after parsing qb64b or qb2. False means do not strip

        Needs either (raw and code and index) or qb64b or qb64 or qb2
        Otherwise raises EmptyMaterialError
        When raw and code provided then validate that code is correct
        for length of raw  and assign .raw
        Else when qb64b or qb64 or qb2 provided extract and assign
        .raw, .code, .index, .ondex.

        """
        if raw is not None:  # raw provided
            if not code:
                raise EmptyMaterialError("Improper initialization need either "
                                         "(raw and code) or qb64b or qb64 or qb2.")
            if not isinstance(raw, (bytes, bytearray)):
                raise TypeError(f"Not a bytes or bytearray, raw={raw}.")

            if code not in self.Sizes:
                raise UnexpectedCodeError(f"Unsupported code={code}.")

            hs, ss, os, fs, ls = self.Sizes[code]  # get sizes for code
            cs = hs + ss  # both hard + soft code size
            ms = ss - os

            if not isinstance(index, int) or index < 0 or index > (64 ** ms - 1):
                raise InvalidVarIndexError(f"Invalid index={index} for code={code}.")

            if isinstance(ondex, int) and os and not (ondex >= 0 and ondex <= (64 ** os - 1)):
                raise InvalidVarIndexError(f"Invalid ondex={ondex} for code={code}.")

            if code in IdxCrtSigDex and ondex is not None:
                raise InvalidVarIndexError(f"Non None ondex={ondex} for code={code}.")

            if code in IdxBthSigDex:
                if ondex is None:  # set default
                    ondex = index  # when not provided make ondex match index
                else:
                    if ondex != index and os == 0:  # must match if os == 0
                        raise InvalidVarIndexError(f"Non matching ondex={ondex}"
                                                   f" and index={index} for "
                                                   f"code={code}.")


            if not fs:  # compute fs from index
                if cs % 4:
                    raise InvalidCodeSizeError(f"Whole code size not multiple of 4 for "
                                               f"variable length material. cs={cs}.")
                if os != 0:
                    raise InvalidCodeSizeError(f"Non-zero other index size for "
                                               f"variable length material. os={os}.")
                fs = (index * 4) + cs

            rawsize = (fs - cs) * 3 // 4

            raw = raw[:rawsize]  # copy rawsize from stream, may be less
            if len(raw) != rawsize:  # forbids shorter
                raise RawMaterialError(f"Not enougth raw bytes for code={code}"
                                       f"and index={index} ,expected {rawsize} "
                                       f"got {len(raw)}.")

            self._code = code
            self._index = index
            self._ondex = ondex
            self._raw = bytes(raw)  # crypto ops require bytes not bytearray

        elif qb64b is not None:
            self._exfil(qb64b)
            if strip:  # assumes bytearray
                del qb64b[:len(self.qb64b)]  # may be variable length fs

        elif qb64 is not None:
            self._exfil(qb64)

        elif qb2 is not None:
            self._bexfil(qb2)
            if strip:  # assumes bytearray
                del qb2[:len(self.qb2)]  # may be variable length fs

        else:
            raise EmptyMaterialError("Improper initialization need either "
                                     "(raw and code and index) or qb64b or "
                                     "qb64 or qb2.")

    @classmethod
    def _rawSize(cls, code):
        """
        Returns expected raw size in bytes for a given code. Not applicable to
        codes with fs = None
        """
        hs, ss, os, fs, ls = cls.Sizes[code]  # get sizes
        return ((fs - (hs + ss)) * 3 // 4)

    @property
    def code(self):
        """
        Returns ._code
        Makes .code read only
        """
        return self._code


    @property
    def name(self):
        """
        Returns:
            name (str): code name for self.code. Used for annotation for
            primitives like Matter

        """
        return self.Names[self.code]

    @property
    def raw(self):
        """
        Returns ._raw
        Makes .raw read only
        """
        return self._raw

    @property
    def index(self):
        """
        Returns ._index
        Makes .index read only
        """
        return self._index

    @property
    def ondex(self):
        """
        Returns ._ondex
        Makes .ondex read only
        """
        return self._ondex

    @property
    def qb64b(self):
        """
        Property qb64b:
        Returns Fully Qualified Base64 Version encoded as bytes
        Assumes self.raw and self.code are correctly populated
        """
        return self._infil()

    @property
    def qb64(self):
        """
        Property qb64:
        Returns Fully Qualified Base64 Version
        Assumes self.raw and self.code are correctly populated
        """
        return self.qb64b.decode("utf-8")

    @property
    def qb2(self):
        """
        Property qb2:
        Returns Fully Qualified Binary Version Bytes
        """
        return self._binfil()

    def _infil(self):
        """
        Returns fully qualified attached sig base64 bytes computed from
        self.raw, self.code and self.index.

        cs = hs + ss
        os = ss - ms (main index size)
        when fs None then size computed & fs = size * 4 + cs

        """
        code = self.code  # codex value chars hard code
        index = self.index  # main index value
        ondex = self.ondex  # other index value
        raw = self.raw  # bytes or bytearray

        ps = (3 - (len(raw) % 3)) % 3  # if lead then same pad size chars & lead size bytes
        hs, ss, os, fs, ls = self.Sizes[code]
        cs = hs + ss
        ms = ss - os

        if not fs:  # compute fs from index
            if cs % 4:
                raise InvalidCodeSizeError(f"Whole code size not multiple of 4 for "
                                           f"variable length material. cs={cs}.")
            if os != 0:
                raise InvalidCodeSizeError(f"Non-zero other index size for "
                                           f"variable length material. os={os}.")
            fs = (index * 4) + cs

        if index < 0 or index > (64 ** ms - 1):
            raise InvalidVarIndexError(f"Invalid index={index} for code={code}.")

        if (isinstance(ondex, int) and os and
                not (ondex >= 0 and ondex <= (64 ** os - 1))):
            raise InvalidVarIndexError(f"Invalid ondex={ondex} for os={os} and "
                                       f"code={code}.")

        # both is hard code + converted index + converted ondex
        both = (f"{code}{intToB64(index, l=ms)}"
                f"{intToB64(ondex if ondex is not None else 0, l=os)}")

        # check valid pad size for whole code size, assumes ls is zero
        if len(both) != cs:
            raise InvalidCodeSizeError("Mismatch code size = {} with table = {}."
                                       .format(cs, len(both)))

        if (cs % 4) != ps - ls:  # adjusted pad given lead bytes
            raise InvalidCodeSizeError(f"Invalid code={both} for converted"
                                       f" raw pad size={ps}.")

        # prepend pad bytes, convert, then replace pad chars with full derivation
        # code including index,
        full = both.encode("utf-8") + encodeB64(bytes([0] * ps) + raw)[ps - ls:]

        if len(full) != fs:  # invalid size
            raise InvalidCodeSizeError(f"Invalid code={both} for raw size={len(raw)}.")

        return full


    def _binfil(self):
        """
        Returns bytes of fully qualified base2 bytes, that is .qb2
        self.code and self.index  converted to Base2 + self.raw left shifted
        with pad bits equivalent of Base64 decode of .qb64 into .qb2
        """
        code = self.code  # codex chars hard code
        index = self.index  # main index value
        ondex = self.ondex  # other index value
        raw = self.raw  # bytes or bytearray

        ps = (3 - (len(raw) % 3)) % 3  # same pad size chars & lead size bytes
        hs, ss, os, fs, ls = self.Sizes[code]
        cs = hs + ss
        ms = ss - os

        if index < 0 or index > (64 ** ss - 1):
            raise InvalidVarIndexError(f"Invalid index={index} for code={code}.")

        if (isinstance(ondex, int) and os and
                not (ondex >= 0 and ondex <= (64 ** os - 1))):
            raise InvalidVarIndexError(f"Invalid ondex={ondex} for os={os} and "
                                       f"code={code}.")

        if not fs:  # compute fs from index
            if cs % 4:
                raise InvalidCodeSizeError(f"Whole code size not multiple of 4 for "
                                           f"variable length material. cs={cs}.")
            if os != 0:
                raise InvalidCodeSizeError(f"Non-zero other index size for "
                                           f"variable length material. os={os}.")
            fs = (index * 4) + cs

        # both is hard code + converted index
        both = (f"{code}{intToB64(index, l=ms)}"
                f"{intToB64(ondex if ondex is not None else 0, l=os)}")

        if len(both) != cs:
            raise InvalidCodeSizeError("Mismatch code size = {} with table = {}."
                                       .format(cs, len(both)))

        if (cs % 4) != ps - ls:  # adjusted pad given lead bytes
                    raise InvalidCodeSizeError(f"Invalid code={both} for converted"
                                               f" raw pad size={ps}.")

        n = sceil(cs * 3 / 4)  # number of b2 bytes to hold b64 code + index
        # convert code both to right align b2 int then left shift in pad bits
        # then convert to bytes
        bcode = (b64ToInt(both) << (2 * (ps - ls))).to_bytes(n, 'big')
        full = bcode + bytes([0] * ls) + raw

        bfs = len(full)  # binary full size
        if bfs % 3 or (bfs * 4 // 3) != fs:  # invalid size
            raise InvalidCodeSizeError(f"Invalid code={both} for raw size={len(raw)}.")

        return full


    def _exfil(self, qb64b):
        """
        Extracts self.code, self.index, and self.raw from qualified base64 bytes qb64b

        cs = hs + ss
        ms = ss - os (main index size)
        when fs None then size computed & fs = size * 4 + cs
        """
        if not qb64b:  # empty need more bytes
            raise ShortageError("Empty material.")

        first = qb64b[:1]  # extract first char code selector
        if hasattr(first, "decode"):
            first = first.decode("utf-8")
        if first not in self.Hards:
            if first[0] == '-':
                raise UnexpectedCountCodeError("Unexpected count code start"
                                               "while extracing Indexer.")
            elif first[0] == '_':
                raise UnexpectedOpCodeError("Unexpected  op code start"
                                            "while extracing Indexer.")
            else:
                raise UnexpectedCodeError(f"Unsupported code start char={first}.")

        hs = self.Hards[first]  # get hard code size
        if len(qb64b) < hs:  # need more bytes
            raise ShortageError(f"Need {hs - len(qb64b)} more characters.")

        hard = qb64b[:hs]  # get hard code
        if hasattr(hard, "decode"):
            hard = hard.decode("utf-8")
        if hard not in self.Sizes:
            raise UnexpectedCodeError(f"Unsupported code ={hard}.")

        hs, ss, os, fs, ls = self.Sizes[hard]  # assumes hs in both tables consistent
        cs = hs + ss  # both hard + soft code size
        ms = ss - os
        # assumes that unit tests on Indexer and IndexerCodex ensure that
        # .Codes and .Sizes are well formed.
        # hs consistent and hs > 0 and ss > 0 and (fs >= hs + ss if fs is not None else True)
        # assumes no variable length indexed codes so fs is not None

        if len(qb64b) < cs:  # need more bytes
            raise ShortageError(f"Need {cs - len(qb64b)} more characters.")

        index = qb64b[hs:hs+ms]  # extract index/size chars
        if hasattr(index, "decode"):
            index = index.decode("utf-8")
        index = b64ToInt(index)  # compute int index

        ondex = qb64b[hs+ms:hs+ms+os]  # extract ondex chars
        if hasattr(ondex, "decode"):
            ondex = ondex.decode("utf-8")

        if hard in IdxCrtSigDex:  # if current sig then ondex from code must be 0
            ondex = b64ToInt(ondex) if os else None  # compute ondex from code
            if ondex:  # not zero or None so error
                raise ValueError(f"Invalid ondex={ondex} for code={hard}.")
            else:
                ondex = None  # zero so set to None when current only
        else:
            ondex = b64ToInt(ondex) if os else index

        # index is index for some codes and variable length for others
        if not fs:  # compute fs from index which means variable length
            if cs % 4:
                raise ValidationError(f"Whole code size not multiple of 4 for "
                                      f"variable length material. cs={cs}.")
            if os != 0:
                raise ValidationError(f"Non-zero other index size for "
                                      f"variable length material. os={os}.")
            fs = (index * 4) + cs

        if len(qb64b) < fs:  # need more bytes
            raise ShortageError(f"Need {fs - len(qb64b)} more chars.")

        qb64b = qb64b[:fs]  # fully qualified primitive code plus material
        if hasattr(qb64b, "encode"):  # only convert extracted chars from stream
            qb64b = qb64b.encode("utf-8")

        # strip off prepended code and append pad characters
        #ps = cs % 4  # pad size ps = cs mod 4, same pad chars and lead bytes
        #base = ps * b'A' + qb64b[cs:]  # replace prepend code with prepad zeros
        #raw = decodeB64(base)[ps+ls:]  # decode and strip off ps+ls prepad bytes

        # check for non-zeroed pad bits or lead bytes
        ps = cs % 4  # code pad size ps = cs mod 4
        pbs = 2 * (ps if ps else ls)  # pad bit size in bits
        if ps:  # ps. IF ps THEN not ls (lead) and vice versa OR not ps and not ls
            base = ps * b'A' + qb64b[cs:]  # replace pre code with prepad chars of zero
            paw = decodeB64(base)  # decode base to leave prepadded raw
            pi = (int.from_bytes(paw[:ps], "big"))  # prepad as int
            if pi & (2 ** pbs - 1 ):  # masked pad bits non-zero
                raise ValueError(f"Non zeroed prepad bits = "
                                 f"{pi & (2 ** pbs - 1 ):<06b} in {qb64b[cs:cs+1]}.")
            raw = paw[ps:]  # strip off ps prepad paw bytes
        else:  # not ps. IF not ps THEN may or may not be ls (lead)
            base = qb64b[cs:]  # strip off code leaving lead chars if any and value
            # decode lead chars + val leaving lead bytes + raw bytes
            # then strip off ls lead bytes leaving raw
            paw = decodeB64(base) # decode base to leave prepadded paw bytes
            li = int.from_bytes(paw[:ls], "big")  # lead as int
            if li:  # pre pad lead bytes must be zero
                if ls == 1:
                    raise ValueError(f"Non zeroed lead byte = 0x{li:02x}.")
                else:
                    raise ValueError(f"Non zeroed lead bytes = 0x{li:04x}.")

            raw = paw[ls:]

        if len(raw) != (len(qb64b) - cs) * 3 // 4:  # exact lengths
            raise ConversionError(f"Improperly qualified material = {qb64b}")

        self._code = hard
        self._index = index
        self._ondex = ondex
        self._raw = raw  # must be bytes for crpto opts and immutable not bytearray



    def _bexfil(self, qb2):
        """
        Extracts self.code, self.index, and self.raw from qualified base2 bytes qb2

        cs = hs + ss
        ms = ss - os (main index size)
        when fs None then size computed & fs = size * 4 + cs
        """
        if not qb2:  # empty need more bytes
            raise ShortageError("Empty material, Need more bytes.")

        first = nabSextets(qb2, 1)  # extract first sextet as code selector
        if first not in self.Bards:
            if first[0] == b'\xf8':  # b64ToB2('-')
                raise UnexpectedCountCodeError("Unexpected count code start"
                                               "while extracing Matter.")
            elif first[0] == b'\xfc':  # b64ToB2('_')
                raise UnexpectedOpCodeError("Unexpected  op code start"
                                            "while extracing Matter.")
            else:
                raise UnexpectedCodeError(f"Unsupported code start sextet={first}.")

        hs = self.Bards[first]  # get code hard size equvalent sextets
        bhs = sceil(hs * 3 / 4)  # bhs is min bytes to hold hs sextets
        if len(qb2) < bhs:  # need more bytes
            raise ShortageError(f"Need {bhs - len(qb2)} more bytes.")

        hard = codeB2ToB64(qb2, hs)  # extract and convert hard part of code
        if hard not in self.Sizes:
            raise UnexpectedCodeError(f"Unsupported code ={hard}.")

        hs, ss, os, fs, ls = self.Sizes[hard]
        cs = hs + ss  # both hs and ss
        ms = ss - os
        # assumes that unit tests on Indexer and IndexerCodex ensure that
        # .Codes and .Sizes are well formed.
        # hs consistent and hs > 0 and ss > 0 and (fs >= hs + ss if fs is not None else True)

        bcs = sceil(cs * 3 / 4)  # bcs is min bytes to hold cs sextets
        if len(qb2) < bcs:  # need more bytes
            raise ShortageError("Need {} more bytes.".format(bcs - len(qb2)))

        both = codeB2ToB64(qb2, cs)  # extract and convert both hard and soft part of code
        index = b64ToInt(both[hs:hs+ms])  # compute index

        if hard in IdxCrtSigDex:  # if current sig then ondex from code must be 0
            ondex = b64ToInt(both[hs+ms:hs+ms+os]) if os else None  # compute ondex from code
            if ondex:  # not zero or None so error
                raise ValueError(f"Invalid ondex={ondex} for code={hard}.")
            else:
                ondex = None  # zero so set to None when current only
        else:
            ondex = b64ToInt(both[hs+ms:hs+ms+os]) if os else index

        if hard in IdxCrtSigDex:  # if current sig then ondex from code must be 0
            if ondex:  # not zero so error
                raise ValueError(f"Invalid ondex={ondex} for code={hard}.")
            else:  # zero so set to None
                ondex = None

        if not fs:  # compute fs from size chars in ss part of code
            if cs % 4:
                raise ValidationError(f"Whole code size not multiple of 4 for "
                                      f"variable length material. cs={cs}.")
            if os != 0:
                raise ValidationError(f"Non-zero other index size for "
                                      f"variable length material. os={os}.")
            fs = (index * 4) + cs

        bfs = sceil(fs * 3 / 4)  # bfs is min bytes to hold fs sextets
        if len(qb2) < bfs:  # need more bytes
            raise ShortageError("Need {} more bytes.".format(bfs - len(qb2)))

        qb2 = qb2[:bfs]  # extract qb2 fully qualified primitive code plus material

        # check for non-zeroed prepad bits or lead bytes
        ps = cs % 4  # code pad size ps = cs mod 4
        pbs = 2 * (ps if ps else ls)  # pad bit size in bits
        if ps:  # ps. IF ps THEN not ls (lead) and vice versa OR not ps and not ls
            # convert last byte of code bytes in which are pad bits to int
            pi = (int.from_bytes(qb2[bcs-1:bcs], "big"))
            if pi & (2 ** pbs - 1 ):  # masked pad bits non-zero
                raise ValueError(f"Non zeroed pad bits = "
                                 f"{pi & (2 ** pbs - 1 ):>08b} in 0x{pi:02x}.")
        else:  # not ps. IF not ps THEN may or may not be ls (lead)
            li = int.from_bytes(qb2[bcs:bcs+ls], "big")  # lead as int
            if li:  # pre pad lead bytes must be zero
                if ls == 1:
                    raise ValueError(f"Non zeroed lead byte = 0x{li:02x}.")
                else:
                    raise ValueError(f"Non zeroed lead bytes = 0x{li:02x}.")


        raw = qb2[(bcs + ls):]  # strip code and leader bytes from qb2 to get raw

        if len(raw) != (len(qb2) - bcs - ls):  # exact lengths
            raise ConversionError(f"Improperly qualified material = {qb2}")

        self._code = hard
        self._index = index
        self._ondex = ondex
        self._raw = bytes(raw)  # must be bytes for crypto ops and not bytearray mutable


class Siger(Indexer):
    """
    Siger is subclass of Indexer, indexed signature material,

    Adds .verfer property which is instance of Verfer that provides
          associated signature verifier.

    See Indexer for inherited attributes and properties:

    Attributes:

    Properties:
        verfer (Verfer): instance if any provides public verification key

    Methods:

    Hidden:
        _verfer (Verfer): value for .verfer property


    """

    def __init__(self, verfer=None, **kwa):
        """Initialze instance

        Parameters:  See Matter for inherted parameters
            verfer (Verfer): instance if any provides public verification key

        """
        super(Siger, self).__init__(**kwa)
        if self.code not in IdxSigDex:
            raise ValidationError("Invalid code = {} for Siger."
                                  "".format(self.code))
        self.verfer = verfer

    @property
    def verfer(self):
        """
        Property verfer:
        Returns Verfer instance
        Assumes ._verfer is correctly assigned
        """
        return self._verfer

    @verfer.setter
    def verfer(self, verfer):
        """ verfer property setter """
        self._verfer = verfer





# -*- encoding: utf-8 -*-
"""
keri.peer.exchanging module

"""
import datetime
import logging
from datetime import timedelta

from hio.help import decking

from .. import help, kering, core
from ..app import habbing
from ..core import eventing, coring, serdering
from ..help import helping
from ..kering import ValidationError, MissingSignatureError

ExchangeMessageTimeWindow = timedelta(seconds=300)

logger = help.ogler.getLogger()


class Exchanger:
    """
     Peer to Peer KERI message Exchanger.
    """

    TimeoutPSE = 10  # seconds to timeout partially signed or delegated escrows

    def __init__(self, hby, handlers, cues=None, delta=ExchangeMessageTimeWindow):
        """ Initialize instance

        Parameters:
            hby (Haberyu): database environment
            handlers(list): list of Handlers capable of responding to exn messages
            cues (Deck):  of Cues i.e. notices of requests needing response
            delta (timedelta): message timeout window
        """

        self.hby = hby
        self.kevers = self.hby.db.kevers
        self.delta = delta
        self.routes = dict()
        self.cues = cues if cues is not None else decking.Deck()  # subclass of deque

        for handler in handlers:
            if handler.resource in self.routes:
                raise ValidationError("unable to register behavior {}, it has already been registered"
                                      "".format(handler.resource))

            self.routes[handler.resource] = handler

    def addHandler(self, handler):
        if handler.resource in self.routes:
            raise ValidationError("unable to register behavior {}, it has already been registered"
                                  "".format(handler.resource))

        self.routes[handler.resource] = handler

    def processEvent(self, serder, tsgs=None, cigars=None, **kwargs):
        """ Process one serder event with attached indexed signatures representing a Peer to Peer exchange message.

        Parameters:
            serder (Serder): instance of event to process
            tsgs (list): tuples (quadruples) of form
                           (prefixer, seqner, diger, [sigers]) where:
                           prefixer is pre of trans endorser
                           seqner is sequence number of trans endorser's est evt for keys for sigs
                           diger is digest of trans endorser's est evt for keys for sigs
                           [sigers] is list of indexed sigs from trans endorser's keys from est evt
            cigars (list): of Cigar instances of attached non-trans sigs

        """
        route = serder.ked["r"]
        sender = serder.ked["i"]
        pathed = kwargs["pathed"] if "pathed" in kwargs else []
        essrs = kwargs["essrs"] if "essrs" in kwargs else []

        behavior = self.routes[route] if route in self.routes else None
        if tsgs:
            for prefixer, seqner, ssaider, sigers in tsgs:  # iterate over each tsg
                if sender != prefixer.qb64:  # sig not by aid
                    raise MissingSignatureError(f"Exchange process: skipped signature not from aid="
                                                f"{sender}, from {prefixer.qb64} on exn msg=\n{serder.pretty()}\n")

                if prefixer.qb64 not in self.kevers or self.kevers[prefixer.qb64].sn < seqner.sn:
                    if self.escrowPSEvent(serder=serder, tsgs=tsgs, pathed=pathed):
                        self.cues.append(dict(kin="query", q=dict(r="logs", pre=prefixer.qb64, sn=seqner.snh)))
                    raise MissingSignatureError(f"Unable to find sender {prefixer.qb64} in kevers"
                                                f" for evt = {serder.ked}.")

                # Verify the signatures are valid and that the signature threshold as of the signing event is met
                tholder, verfers = self.hby.db.resolveVerifiers(pre=prefixer.qb64, sn=seqner.sn, dig=ssaider.qb64)
                _, indices = eventing.verifySigs(serder.raw, sigers, verfers)

                if not tholder.satisfy(indices):  # We still don't have all the sigers, need to escrow
                    if self.escrowPSEvent(serder=serder, tsgs=tsgs, pathed=pathed):
                        self.cues.append(dict(kin="query", q=dict(r="logs", pre=prefixer.qb64, sn=seqner.snh)))
                    raise MissingSignatureError(f"Not enough signatures in  {indices}"
                                                f" for evt = {serder.ked}.")

        elif cigars:
            for cigar in cigars:
                if sender != cigar.verfer.qb64:  # cig not by aid
                    raise MissingSignatureError(" process: skipped cig not from aid="
                                                "%s on exn msg=\n%s\n", sender, serder.pretty())

                if not cigar.verfer.verify(cigar.raw, serder.raw):  # cig not verify
                    raise MissingSignatureError("Failure satisfying exn on cigs for {}"
                                                " for evt = {}.".format(cigar,
                                                                        serder.ked))
        else:
            self.escrowPSEvent(serder=serder, tsgs=[], pathed=pathed)
            raise MissingSignatureError("Failure satisfying exn, no cigs or sigs"
                                        " for evt = {}.".format(serder.ked))

        e = coring.Pather(path=["e"])

        kwargs = dict()
        attachments = []
        for p in pathed:
            pattach = bytearray(p)
            pather = coring.Pather(qb64b=pattach, strip=True)
            if pather.startswith(e):
                np = pather.strip(e)
                attachments.append((np, pattach))

        kwargs["attachments"] = attachments
        if essrs:
            kwargs["essr"] = b''.join([texter.raw for texter in essrs])

        if isinstance(serder.seals, str):
            if 'essr' not in kwargs:
                raise ValidationError("at least one essr attachment is required")

            essr = kwargs['essr']
            dig = serder.seals
            diger = coring.Diger(qb64=dig)
            if not diger.verify(ser=essr):
                raise ValidationError(f"essr diger={diger.qb64} is invalid against content")

        # Perform behavior specific verification, think IPEX chaining requirements
        try:
            if not behavior.verify(serder=serder, **kwargs):
                logger.info(f"exn event for route {route} failed behavior verfication.  said={serder.said}")
                logger.debug(f"event=\n{serder.pretty()}\n")
                return

        except AttributeError:
            logger.info(f"Behavior for {route} missing or does not have verify for said={serder.said}")
            logger.debug(f"event=\n{serder.pretty()}\n")

        # Always persist events
        self.logEvent(serder, pathed, tsgs, cigars, essrs)
        self.cues.append(dict(kin="saved", said=serder.said))

        # Execute any behavior specific handling, not sure if this should be different than verify
        try:
            behavior.handle(serder=serder, **kwargs)
        except AttributeError:
            logger.info(f"Behavior for {route} missing or does not have handle for said={serder.said}")
            logger.debug(f"event=\n{serder.pretty()}\n")

    def processEscrow(self):
        """ Process all escrows for `exn` messages

        """
        self.processEscrowPartialSigned()

    def escrowPSEvent(self, serder, tsgs, pathed):
        """ Escrow event that does not have enough signatures.

        Parameters:
            serder (Serder): instance of event
            tsgs (list): quadlet of prefixer seqner, saider, sigers
            pathed (list): list of bytes of attached paths

        """
        dig = serder.said
        for prefixer, seqner, ssaider, sigers in tsgs:  # iterate over each tsg
            quadkeys = (serder.said, prefixer.qb64, f"{seqner.sn:032x}", ssaider.qb64)
            for siger in sigers:
                self.hby.db.esigs.add(keys=quadkeys, val=siger)

        self.hby.db.epsd.put(keys=(dig,), val=coring.Dater())
        self.hby.db.epath.pin(keys=(dig,), vals=[bytes(p) for p in pathed])
        return self.hby.db.epse.put(keys=(dig,), val=serder)

    def processEscrowPartialSigned(self):
        """ Process escrow of partially signed messages """
        for (dig,), serder in self.hby.db.epse.getItemIter():
            try:
                tsgs = []
                klases = (coring.Prefixer, coring.Seqner, coring.Saider)
                args = ("qb64", "snh", "qb64")
                sigers = []

                dtnow = helping.nowUTC()
                dater = self.hby.db.epsd.get(keys=(dig,))
                if dater is None:
                    raise ValidationError("Missing exn escrowed event datetime "
                                          f"at dig = {dig}.")

                dte = dater.datetime
                if (dtnow - dte) > datetime.timedelta(seconds=self.TimeoutPSE):
                    # escrow stale so raise ValidationError which unescrows below
                    raise ValidationError("Stale exn event escrow "
                                          f"at dig = {dig}.")

                old = None  # empty keys
                for keys, siger in self.hby.db.esigs.getItemIter(keys=(dig, "")):
                    quad = keys[1:]
                    if quad != old:  # new tsg
                        if sigers:  # append tsg made for old and sigers
                            prefixer, seqner, saider = helping.klasify(sers=old, klases=klases, args=args)

                            tsgs.append((prefixer, seqner, saider, sigers))
                            sigers = []
                        old = quad
                    sigers.append(siger)
                if sigers and old:
                    prefixer, seqner, saider = helping.klasify(sers=old, klases=klases, args=args)
                    tsgs.append((prefixer, seqner, saider, sigers))

                pathed = [bytearray(p.encode("utf-8")) for p in self.hby.db.epath.get(keys=(dig,))]
                essrs = [texter for texter in self.hby.db.essrs.get(keys=(dig,))]

                kwargs = dict()
                if essrs:
                    kwargs["essrs"] = essrs
                self.processEvent(serder=serder, tsgs=tsgs, pathed=pathed, **kwargs)

            except MissingSignatureError as ex:
                if logger.isEnabledFor(logging.DEBUG):
                    logger.info("Exchange partially signed unescrow failed: %s", ex.args[0])
                else:
                    logger.info("Exchange partially signed failed: %s", ex.args[0])
            except Exception as ex:
                self.hby.db.epse.rem(dig)
                self.hby.db.epsd.rem(dig)
                self.hby.db.esigs.rem(dig)
                if logger.isEnabledFor(logging.DEBUG):
                    logger.exception("Exchange partially signed unescrowed: %s", ex.args[0])
                else:
                    logger.error("Exchange partially signed unescrowed: %s", ex.args[0])
            else:
                self.hby.db.epse.rem(dig)
                self.hby.db.esigs.rem(dig)
                logger.info("Exchanger unescrow succeeded in valid exchange: "
                            "creder=%s", serder.said)
                logger.debug(f"event=\n{serder.pretty()}\n")

    def logEvent(self, serder, pathed=None, tsgs=None, cigars=None, essrs=None):
        dig = serder.said
        pdig = serder.ked['p']
        pathed = pathed or []
        tsgs = tsgs or []
        cigars = cigars or []
        essrs = essrs or []

        for prefixer, seqner, ssaider, sigers in tsgs:  # iterate over each tsg
            quadkeys = (serder.said, prefixer.qb64, f"{seqner.sn:032x}", ssaider.qb64)
            for siger in sigers:
                self.hby.db.esigs.add(keys=quadkeys, val=siger)
        for cigar in cigars:
            self.hby.db.ecigs.add(keys=(dig,), val=(cigar.verfer, cigar))

        saider = coring.Saider(qb64=serder.said)
        self.hby.db.epath.pin(keys=(dig,), vals=[bytes(p) for p in pathed])
        for texter in essrs:
            self.hby.db.essrs.add(keys=(dig,), val=texter)
        if pdig:
            self.hby.db.erpy.pin(keys=(pdig,), val=saider)

        self.hby.db.exns.put(keys=(dig,), val=serder)

    def lead(self, hab, said):
        """ Determines is current member represented by hab is the lead of an exn message

        Lead is the signer of the exn with the lowest signing index

        Parameters:
            hab (Hab): Habitat for sending of exchange message represented by SAID
            said (str): qb64 SAID of exchange message

        Returns:
            bool: True means hab is the lead

        """
        if not isinstance(hab, habbing.GroupHab):
            return True

        keys = [verfer.qb64 for verfer in hab.kever.verfers]
        tsgs = eventing.fetchTsgs(self.hby.db.esigs, coring.Saider(qb64=said))
        if not tsgs:  # otherwise it contains a list of sigs
            return False

        (_, _, _, sigers) = tsgs[0]
        windex = min([siger.index for siger in sigers])

        # True if Elected to send an EXN to its recipient
        return hab.mhab.kever.verfers[0].qb64 == keys[windex]

    def complete(self, said):
        """

        Args:
            said (str): qb64 said of exchange message to check status

        Returns:
            bool: True means exchange message is has been saved
        """
        serder = self.hby.db.exns.get(keys=(said,))
        if not serder:
            return False
        else:
            if serder.said != said:
                raise kering.ValidationError(f"invalid exchange escrowed event {serder.said}-{said}")

        return True


def exchange(route,
             sender,
             payload=None,
             diger=None,
             recipient=None,
             date=None,
             dig=None,
             modifiers=None,
             embeds=None,
             version=coring.Version,
             kind=coring.Kinds.json):
    """ Create an `exn` message with the specified route and payload

    Parameters:
        route (str): to destination route of the message
        payload (list | dict): body of message to deliver to route
        diger (Diger): qb64 digest of payload
        sender (str): qb64 AID of sender of the exn
        recipient (str) optional qb64 AID recipient of exn
        date (str): Iso8601 formatted date string to use for this request
        dig (str) qb64 SAID of previous event if any
        modifiers (dict): equivalent of query string of uri, modifiers for the request that are not
                         part of the payload
        embeds (dict): named embeded KERI event CESR stream with attachments
        version (Version): is Version instance
        kind (Serials): is serialization kind

    """
    vs = coring.versify(version=version, kind=kind, size=0)
    ilk = eventing.Ilks.exn
    dt = date if date is not None else helping.nowIso8601()
    p = dig if dig is not None else ""
    rp = recipient if recipient is not None else ""
    embeds = embeds if embeds is not None else {}

    e = dict()
    end = bytearray()
    for label, msg in embeds.items():
        serder = coring.Sadder(raw=msg)
        e[label] = serder.ked
        atc = bytes(msg[serder.size:])
        if not atc:
            continue

        pathed = bytearray()
        pather = coring.Pather(path=["e", label])
        pathed.extend(pather.qb64b)
        pathed.extend(atc)
        if len(pathed) // 4 < 4096:
            end.extend(core.Counter(core.Codens.PathedMaterialGroup,
                                      count=(len(pathed) // 4),
                                      gvrsn=kering.Vrsn_1_0).qb64b)
        else:
            end.extend(core.Counter(core.Codens.BigPathedMaterialGroup,
                                      count=(len(pathed) // 4),
                                      gvrsn=kering.Vrsn_1_0).qb64b)
        end.extend(pathed)

    if e:
        e["d"] = ""
        _, e = coring.Saider.saidify(sad=e, label=coring.Saids.d)

    modifiers = modifiers if modifiers is not None else {}

    if diger is None:
        attrs = dict()

        if recipient is not None:
            attrs['i'] = recipient

        attrs |= payload

    else:
        attrs = diger.qb64

    # Attr field 'a' can be either a said or a nested block and the fields
    # of the nested block can be saids of further nested block or nested blocks
    ked = dict(v=vs,
               t=ilk,
               d="",
               i=sender,
               rp=rp,
               p=p,
               dt=dt,
               r=route,
               q=modifiers if modifiers is not None else {},  # q field required
               a=attrs,
               e=e)

    _, ked = coring.Saider.saidify(sad=ked)
    return serdering.SerderKERI(sad=ked), end  # return serialized ked


def cloneMessage(hby, said):
    """ Load and verify signatures on message exn

    Parameters:
        hby (Habery): database environment from which to clone message
        said (str): qb64 SAID of message exn to load

    Returns:
        tuple: (serder, list) of message exn and pathed signatures on embedded attachments

    """
    exn = hby.db.exns.get(keys=(said,))
    if exn is None:
        return None, None

    verify(hby=hby, serder=exn)

    pathed = dict()
    e = coring.Pather(path=["e"])
    for p in hby.db.epath.get(keys=(exn.said,)):
        pb = bytearray(p.encode("utf-8"))
        pather = coring.Pather(qb64b=pb, strip=True)
        if pather.startswith(e):
            np = pather.strip(e)
            nesting(np.path, pathed, pb)

    return exn, pathed


def serializeMessage(hby, said, pipelined=False):
    atc = bytearray()

    exn = hby.db.exns.get(keys=(said,))
    if exn is None:
        return None, None

    atc.extend(exn.raw)

    tsgs, cigars = verify(hby=hby, serder=exn)

    if len(tsgs) > 0:
        for (prefixer, seqner, saider, sigers) in tsgs:
            atc.extend(core.Counter(core.Codens.TransIdxSigGroups, count=1,
                                    gvrsn=kering.Vrsn_1_0).qb64b)
            atc.extend(prefixer.qb64b)
            atc.extend(seqner.qb64b)
            atc.extend(saider.qb64b)

            atc.extend(core.Counter(core.Codens.ControllerIdxSigs, count=len(sigers),
                                    gvrsn=kering.Vrsn_1_0).qb64b)
            for siger in sigers:
                atc.extend(siger.qb64b)

    if len(cigars) > 0:
        atc.extend(core.Counter(core.Codens.NonTransReceiptCouples,
                                count=len(cigars), gvrsn=kering.Vrsn_1_0).qb64b)
        for cigar in cigars:
            if cigar.verfer.code not in coring.NonTransDex:
                raise ValueError("Attempt to use tranferable prefix={} for "
                                 "receipt.".format(cigar.verfer.qb64))
            atc.extend(cigar.verfer.qb64b)
            atc.extend(cigar.qb64b)

    # Smash the pathed components on the end
    for p in hby.db.epath.get(keys=(exn.said,)):
        atc.extend(core.Counter(core.Codens.PathedMaterialGroup,
                                  count=(len(p) // 4), gvrsn=kering.Vrsn_1_0).qb64b)
        atc.extend(p.encode("utf-8"))

    msg = bytearray()

    if pipelined:
        if len(atc) % 4:
            raise ValueError("Invalid attachments size={}, nonintegral"
                             " quadlets.".format(len(atc)))
        msg.extend(core.Counter(core.Codens.AttachmentGroup,
                                  count=(len(atc) // 4), gvrsn=kering.Vrsn_1_0).qb64b)

    msg.extend(atc)
    return msg


def nesting(paths, acc, val):
    if len(paths) == 0:
        return val
    else:
        first_value = paths[0]
        nacc = dict()
        acc[first_value] = nesting(paths[1:], nacc, val)
        return acc


def verify(hby, serder):
    """  Verify that the signatures in the database are valid for the provided exn

    Parameters:
        hby (Habery): database environment from which to verify message
        serder (Serder): exn serder to load and verify signatures for

    Returns:
        bool: True means threshold satisfyig signatures were loaded and verified successfully

    """
    tsgs = []
    klases = (coring.Prefixer, coring.Seqner, coring.Saider)
    args = ("qb64", "snh", "qb64")
    sigers = []
    old = None  # empty keys
    for keys, siger in hby.db.esigs.getItemIter(keys=(serder.said, "")):
        quad = keys[1:]
        if quad != old:  # new tsg
            if sigers:  # append tsg made for old and sigers
                prefixer, seqner, saider = helping.klasify(sers=old, klases=klases, args=args)

                tsgs.append((prefixer, seqner, saider, sigers))
                sigers = []
            old = quad
        sigers.append(siger)
    if sigers and old:
        prefixer, seqner, saider = helping.klasify(sers=old, klases=klases, args=args)
        tsgs.append((prefixer, seqner, saider, sigers))

    accepted = False
    for prefixer, seqner, ssaider, sigers in tsgs:
        if prefixer.qb64 not in hby.kevers or hby.kevers[prefixer.qb64].sn < seqner.sn:
            raise MissingSignatureError(f"Unable to find sender {prefixer.qb64} in kevers"
                                        f" for evt = {serder.ked}.")

        # Verify the signatures are valid and that the signature threshold as of the signing event is met
        tholder, verfers = hby.db.resolveVerifiers(pre=prefixer.qb64, sn=seqner.sn, dig=ssaider.qb64)
        _, indices = eventing.verifySigs(serder.raw, sigers, verfers)

        if not tholder.satisfy(indices):  # We still don't have all the sigers, need to escrow
            raise MissingSignatureError(f"Not enough signatures in  {indices}"
                                        f" for evt = {serder.ked}.")
        accepted = True

    cigars = hby.db.ecigs.get(keys=(serder.said,))
    for cigar in cigars:
        if not cigar.verfer.verify(cigar.raw, serder.raw):  # cig not verify
            raise MissingSignatureError("Failure satisfying exn on cigs for {}"
                                        " for evt = {}.".format(cigar,
                                                                serder.ked))
        accepted = True

    if not accepted:
        raise MissingSignatureError(f"No valid signatures stored for evt = {serder.ked}")

    return tsgs, cigars



# -*- encoding: utf-8 -*-
"""
keri.app..cli.common.incepting module

"kli incept" command configuration options
"""


def addInceptingArgs(parser):
    """
    Add command line arguments for each of the properties in InceptOptions
    """
    parser.add_argument('--transferable', '-tf', action="store_true",
                        help='Whether the prefix is transferable or non-transferable')
    parser.add_argument('--wits',         '-w', default=[], required=False, action="append", metavar="<prefix>",
                        help='New set of witnesses, replaces all existing witnesses.  Can appear multiple times')
    parser.add_argument('--toad',         '-t', default=None, required=False, type=int,
                        help='int or str hex of witness threshold (threshold of accountable duplicity)',)
    parser.add_argument('--icount',       '-ic', default=None, required=False,
                        help='incepting key count for number of keys used for inception')
    parser.add_argument('--isith',        '-s', default=None, required=False,
                        help='signing threshold for the inception event')
    parser.add_argument('--ncount',       '-nc', default=None, required=False,
                        help='next key count for number of next keys used on first rotation')
    parser.add_argument('--nsith',        '-x', default=None, required=False,
                        help='signing threshold for the next rotation event',)
    parser.add_argument('--est-only',     '-e', type=bool, default=None,
                        help='only allow establishment events in KEL for this prefix')
    parser.add_argument('--data',         '-d', default=None, required=False, action="store",
                        help='Anchor data, \'@\' allowed',)
    parser.add_argument('--delpre',       '-di', default=None, required=False, action="store",
                        help='Delegator AID for delegated identfiers',)




# -*- coding: utf-8 -*-
"""
keri.core.streaming module

Provides support for Streamer and Annotater
"""


from typing import NamedTuple
from collections import namedtuple

from .. import kering
from ..kering import sniff, Colds, Ilks

from ..help.helping import intToB64



from .. import help

from . import coring
from .coring import (Matter, Verser, Ilker, Diger, Prefixer, Number, Tholder,
                     Verfer, Traitor)

from . import counting
from .counting import Counter

from . import structing
from .structing import Sealer

from . import serdering
from .serdering import Serder


def annot(ims):
    """Annotate CESR stream

    Returns:
        annotation (str):  annotation of input CESR stream

    Parameters:
        ims (str | bytes | bytearray | memoryview): CESR incoming message stream
           as qb64  (maybe qb2)

    """
    oms = bytearray()
    indent = 0


    if not isinstance(ims, bytearray):  # going to strip
        ims = bytearray(ims)  # so make bytearray copy, converts str to bytearray

    while ims:  # right now just for KERI event messages
        cold = sniff(ims)  # check for spurious counters at front of stream

        if cold in Colds.txt:
            val = Counter(qb64b=ims, strip=True)
            oms.extend(f"{' ' * indent * 2}{val.qb64} # Key Event Counter "
                       f"{val.name} count={val.count} quadlets\n".encode())
            indent += 1

            # should grab fill message frame here so can verify its consumed
            # when done annotating

            # version
            val = Verser(qb64b=ims, strip=True)
            versage = val.versage
            oms.extend(f"{' ' * indent * 2}{val.qb64} # 'v' version  Verser {val.name} "
                       f"proto={versage.proto} vrsn={versage.vrsn.major}."
                       f"{versage.vrsn.minor:02}\n".encode())
            # ilk
            ilker = Ilker(qb64b=ims, strip=True)
            ilk = ilker.ilk
            oms.extend(f"{' ' * indent * 2}{ilker.qb64} # 't' message type Ilker "
                       f"{ilker.name} Ilk={ilker.ilk}\n".encode())



            if ilk in (Ilks.icp, Ilks.ixn, Ilks.rot, Ilks.dip, Ilks.drt):
                # said
                val = Diger(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'd' SAID Diger "
                           f"{val.name} \n".encode())

                # aid pre
                val = Prefixer(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'i' AID Prefixer "
                           f"{val.name} \n".encode())

                # sn
                val = Number(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 's' Number "
                           f"{val.name} sn={val.sn}\n".encode())

            if ilk in (Ilks.icp, Ilks.dip):  # inception
                # v='', t='', d='', i='', s='0', kt='0',k=[], nt='0', n=[], bt='0', b=[], c=[], a=[]

                # Signing key threshold
                val = Tholder(limen=ims, strip=True)  # add qb64 and qb2 to tholder as aliases for limen and limen.decode()
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'kt' "
                           f"Tholder signing threshold={val.sith}\n".encode())
                # Signing key list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'k' Signing Key "
                           f"List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Verfer(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # key Verfer "
                               f"{val.name}\n".encode())
                indent -= 1
                # Rotation key threshold
                val = Tholder(limen=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'nt' "
                           f"Tholder rotation threshold={val.sith}\n".encode())
                # Next key digest list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'n' Rotation Key "
                           f"Digest List Counter {val.name} count={val.count} "
                           f"quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Diger(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # key digest Diger"
                               f" {val.name}\n".encode())
                indent -= 1
                # Witness Backer threshold
                val = Tholder(limen=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'bt' "
                           f"Tholder Backer (witness) threshold={val.sith}\n".encode())
                # Witness Backer list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'b' Backer (witness)"
                           f"List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Prefixer(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # AID Prefixer "
                               f"{val.name}\n".encode())
                indent -= 1
                # Config Trait List
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'c' Config Trait "
                           f"List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Traitor(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # trait Traitor "
                               f"{val.name} trait={val.trait}\n".encode())
                indent -= 1


            elif ilk == Ilks.ixn:  # Interaction
                # v='', t='',d='', i='', s='0', p='', a=[]
                # Prior said
                val = Diger(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'p' prior SAID Diger "
                           f"{val.name} \n".encode())



            elif ilk in (Ilks.rot, Ilks.drt):  # Rotation
                # v='', t='',d='', i='', s='0', p='', kt='0',k=[], nt='0', n=[], bt='0', br=[], ba=[], c=[], a=[]
                # Prior said
                val = Diger(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'p' prior SAID Diger "
                           f"{val.name} \n".encode())

                # Signing key threshold
                val = Tholder(limen=ims, strip=True)  # add qb64 and qb2 to tholder as aliases for limen and limen.decode()
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'kt' "
                           f"Tholder signing threshold={val.sith}\n".encode())
                # Signing key list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'k' Signing Key "
                           f"List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Verfer(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # key Verfer "
                               f"{val.name}\n".encode())
                indent -= 1
                # Rotation key threshold
                val = Tholder(limen=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'nt' "
                           f"Tholder rotation threshold={val.sith}\n".encode())
                # Next key digest list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'n' Rotation Key "
                           f"Digest List Counter {val.name} count={val.count} "
                           f"quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Diger(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # key digest Diger"
                               f" {val.name}\n".encode())
                indent -= 1
                # Witness Backer threshold
                val = Tholder(limen=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.limen.decode()} # 'bt' "
                           f"Tholder Backer (witness) threshold={val.sith}\n".encode())

                # Witness Backer Cut list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'b' Backer (witness)"
                           f"Cut List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Prefixer(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # AID Prefixer "
                               f"{val.name}\n".encode())
                indent -= 1

                # Witness Backer Add list
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'b' Backer (witness)"
                           f"Add List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Prefixer(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # AID Prefixer "
                               f"{val.name}\n".encode())
                indent -= 1

                # Config Trait List
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'c' Config Trait "
                           f"List Counter {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Traitor(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # trait Traitor "
                               f"{val.name} trait={val.trait}\n".encode())
                indent -= 1


            else:
                raise kering.IlkError(f"Unexpected message type={ilk}")

            if ilk in (Ilks.icp, Ilks.ixn, Ilks.rot, Ilks.dip, Ilks.drt):
                # Seal (anchor) List
                val = Counter(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'a' Seal List Counter"
                           f" {val.name} count={val.count} quadlets\n".encode())
                indent += 1
                frame = ims[:val.count*4]  # extract frame from ims
                del ims[:val.count*4]  # strip frame
                while frame:
                    val = Counter(qb64b=frame, strip=True)
                    oms.extend(f"{' ' * indent * 2}{val.qb64} # Seal Counter "
                               f"{val.name} count={val.count} quadlets\n".encode())
                    indent += 1
                    subframe = frame[:val.count*4]
                    del frame[:val.count*4]  # strip subframe
                    clan = Sealer.Clans[Serder.CodeClans[val.code]]
                    while subframe:
                        val = Sealer(clan=clan, qb64=subframe, strip=True)  # need to add qb64 parameter to structor
                        oms.extend(f"{' ' * indent * 2}{val.qb64}# seal Sealer {val.name}\n".encode())
                        indent += 1
                        for i, t in enumerate(val.crew._asdict().items()):
                            oms.extend(f"{' ' * indent * 2}#  '{t[0]}' = "
                                   f"{t[1]}\n".encode())
                        indent -= 1
                    indent -= 1
                indent -= 1

            if ilk == Ilks.dip:
                # delegator aid delpre
                val = Prefixer(qb64b=ims, strip=True)
                oms.extend(f"{' ' * indent * 2}{val.qb64} # 'di' Delegator AID Prefixer "
                           f"{val.name} \n".encode())

            if ims:
                raise kering.ExtractionError(f"Unexpected remaining bytes in stream.")

        elif cold in Colds.bny:
            pass

        else:
            raise kering.ColdStartError("Expecting stream tritet={}"
                                        "".format(cold))

    return oms.decode()  # return unicode string

def denot(ams):
    """De-annotate CESR stream

    Returns:
        dms (bytes):  deannotation of input annotated CESR message stream

    Parameters:
        ams (str):  CESR annotated message stream text
    """
    dms = bytearray()  # deannotated message stream
    lines = ams.splitlines()
    for line in lines:
        line = line.strip()
        front, sep, back = line.partition('#') # finde comment if any
        front = front.strip()  # non-commented portion strip white space
        if front:
            dms.extend(front.encode())

    return bytes(dms)


class Streamer:
    """
    Streamer is CESR sniffable stream class


    Has the following public properties:

    Properties:
        stream (bytearray): sniffable CESR stream


    Methods:


    Hidden:
        _verify() -> bool



    """

    def __init__(self, stream, verify=False):
        """Initialize instance
        Holds sniffable CESR stream as byte like string
        either (bytes, bytearray, or memoryview)


        Parameters:
            stream (str | bytes | bytearray | memoryview): sniffable CESR stream
            verify (bool): When True raise error if .stream is not sniffable.


        """
        if hasattr(stream, "encode"):
            stream = bytearray(stream.encode())  # convert str to bytearray
        if not isinstance(stream, (bytes, bytearray, memoryview)):
            raise kering.InvalidTypeError(f"Invalid stream type, not byteable.")

        self._stream = stream


    def _verify(self):
        """Returns True if .stream is sniffable, False otherwise

        Returns:
            sniffable (bool): True when .stream is sniffable.
                                  False otherwise.
        Only works for ver 2 CESR because need for all count codes to be
        pipelineable in order to simply parse stream

        """
        return False


    @property
    def stream(self):
        """stream property getter
        """
        return self._stream


    @property
    def text(self):
        """expanded stream where all primitives and groups in stream are
        individually expanded to qb64.
        Requires parsing full depth to ensure expanded consistently.
        Returns:
           stream (bytes): expanded text qb64 version of stream

        Only works for ver 2 CESR because need for all count codes to be
        pipelineable in order to simply parse and expand stream

        """
        return self._stream

    @property
    def binary(self):
        """compacted stream where all primitives and groups in stream are
        individually compacted to qb2.
        Requires parsing full depth to ensure compacted consistently
        Returns:
           stream (bytes): compacted binary qb2 version of stream

        Only works for ver 2 CESR because need for all count codes to be
        pipelineable in order to simply parse and compact stream

        """
        return self._stream

    @property
    def texter(self):
        """stream as Texter instance.
            Texter(text=self.stream)
        Returns:
           texter (Texter): Texter primitive of stream suitable wrapping

        """
        return self._stream

    @property
    def bexter(self):
        """stream as Bexter instance.
        Bexter of expanded text version of stream.
        First expand  to text which requires parsing then create bexter
            Bexter(bext=self.text)
        Because sniffable stream MUST NOT start with 'A' then there is no
        length ambiguity. The only tritet collison of 'A' is with '-' but the
        remaining 5 bits are guaranteed to always be different. So bexter must
        check not just the starting tritet but the full starting byte to ensure
        not 'A' as first byte.

        Requires parsing to ensure qb64
        Returns:
           bexter (Bexter): Bexter primitive of stream suitable wrapping

        """
        return self._stream








# -*- encoding: utf-8 -*-
"""
KERI
keri.kli.commands module

"""
import argparse

from hio import help
from hio.base import doing

from keri.app import connecting, habbing, forwarding
from keri.app.cli.common import existing
from keri.core import serdering
from keri.kering import Roles

logger = help.ogler.getLogger()

parser = argparse.ArgumentParser(description='Add AID or Alias to list of AIDs for a watcher to watch')
parser.set_defaults(handler=lambda args: add(args),
                    transferable=True)
parser.add_argument('--name', '-n', help='keystore name and file location of KERI keystore', required=True)
parser.add_argument('--alias', '-a', help='human readable alias for the identifier to whom the credential was issued',
                    required=True)
parser.add_argument('--base', '-b', help='additional optional prefix to file location of KERI keystore',
                    required=False, default="")
parser.add_argument('--passcode', '-p', help='22 character encryption passcode for keystore (is not saved)',
                    dest="bran", default=None)  # passcode => bran
parser.add_argument("--watcher", '-w', help="the watcher AID or alias to add to", required=True)
parser.add_argument("--watched", '-W', help="the watched AID or alias to add")


def add(args):
    """ Command line handler for adding an aid to a watcher's list of AIds to watch

    Parameters:
        args(Namespace): parsed command line arguments

    """

    ed = AddDoer(name=args.name,
                 alias=args.alias,
                 base=args.base,
                 bran=args.bran,
                 watcher=args.watcher,
                 watched=args.watched)
    return [ed]


class AddDoer(doing.DoDoer):

    def __init__(self, name, alias, base, bran, watcher, watched):
        self.hby = existing.setupHby(name=name, base=base, bran=bran)
        self.hab = self.hby.habByName(alias)
        self.org = connecting.Organizer(hby=self.hby)

        wat = None
        if watcher in self.hby.kevers:
            wat = watcher
        else:
            contacts = self.org.find("alias", watcher)
            for contact in contacts:
                if contact['alias'] == watcher:
                    wat = contact['id']

        if not wat:
            raise ValueError(f"unknown watcher {watcher}")

        watd = None
        if watched in self.hby.kevers:
            watd = watched
        else:
            contacts = self.org.find("alias", watched)
            for contact in contacts:
                if contact['alias'] == watched:
                    watd = contact['id']

        if not watd:
            raise ValueError(f"unknown watched {watched}")

        self.watcher = wat
        self.watched = watd

        self.oobi = None
        for (key,), obr in self.hby.db.roobi.getItemIter():
            if obr.cid == watd:
                self.oobi = key

        if not self.oobi:
            raise ValueError(f"no valid oobi for watched {self.watched}")

        doers = [doing.doify(self.addDo)]

        super(AddDoer, self).__init__(doers=doers)

    def addDo(self, tymth, tock=0.0):
        """ Add an AID to a watcher's list of AIDs to watch

        Parameters:
            tymth (function): injected function wrapper closure returned by .tymen() of
                Tymist instance. Calling tymth() returns associated Tymist .tyme.
            tock (float): injected initial tock value

        Returns:  doifiable Doist compatible generator method

        """
        # enter context
        self.wind(tymth)
        self.tock = tock
        _ = (yield self.tock)

        if isinstance(self.hab, habbing.GroupHab):
            raise ValueError("watchers for multisig AIDs not currently supported")

        ender = self.hab.db.ends.get(keys=(self.hab.pre, Roles.watcher, self.watcher))
        if not ender or not ender.allowed:
            msg = self.hab.reply(route="/end/role/add",
                                 data=dict(cid=self.hab.pre, role=Roles.watcher, eid=self.watcher))
            self.hab.psr.parseOne(ims=msg)

        postman = forwarding.StreamPoster(hby=self.hby, hab=self.hab, recp=self.watcher, topic="reply")
        for msg in self.hab.db.cloneDelegation(self.hab.kever):
            serder = serdering.SerderKERI(raw=msg)
            postman.send(serder=serder, attachment=msg[serder.size:])

        for msg in self.hab.db.clonePreIter(pre=self.hab.pre):
            serder = serdering.SerderKERI(raw=msg)
            postman.send(serder=serder, attachment=msg[serder.size:])

        data = dict(cid=self.hab.pre,
                    oid=self.watched,
                    oobi=self.oobi)

        route = f"/watcher/{self.watcher}/add"
        msg = self.hab.reply(route=route, data=data)
        self.hab.psr.parseOne(ims=bytes(msg))
        rpy = serdering.SerderKERI(raw=msg)
        postman.send(serder=rpy, attachment=msg[rpy.size:])

        doer = doing.DoDoer(doers=postman.deliver())
        self.extend([doer])

        while not doer.done:
            yield self.tock

        print(f"Request to add {self.watched} to watcher {self.watcher} submitted.")
